"""
Tests for the thinking utility functions.
"""

import pytest
from unittest.mock import MagicMock, patch

# Mock ChromaMcpClient since we don't need the actual implementation for tests
with patch("chroma_mcp_thinking.utils.ChromaMcpClient"):
    with patch("chroma_mcp_thinking.thinking_session.ClientSession") as MockMcpClient:
        from chroma_mcp_thinking.utils import record_thought_chain, find_thoughts_across_sessions, create_thought_branch


@pytest.fixture
def mock_client():
    """Create a mock ChromaMcpClient."""
    client = MagicMock()
    client.mcp_chroma_dev_chroma_sequential_thinking.return_value = {"session_id": "test-session-id"}
    client.mcp_chroma_dev_chroma_find_similar_thoughts.return_value = {
        "similar_thoughts": [
            {
                "metadata": {"session_id": "test-session-id", "thought_number": 1},
                "document": "Test thought",
                "distance": 0.5,
            }
        ]
    }
    client.mcp_chroma_dev_chroma_get_session_summary.return_value = {
        "thoughts": [{"metadata": {"thought_number": 1, "session_id": "test-session-id"}, "document": "Test thought"}]
    }
    return client


@patch("chroma_mcp_thinking.utils.ThinkingSession")
def test_record_thought_chain(mock_thinking_session_class, mock_client):
    """Test recording a thought chain."""
    # Setup mock session
    mock_session = MagicMock()
    mock_session.session_id = "test-session-id"
    mock_session.get_session_summary.return_value = {"thoughts": [{"document": "Test thought"}]}
    mock_thinking_session_class.return_value = mock_session

    # Call the function
    thoughts = ["Thought 1", "Thought 2", "Thought 3"]
    metadata = {"domain": "testing", "tags": ["test", "example"]}
    result = record_thought_chain(
        thoughts=thoughts, session_id="test-session-id", metadata=metadata, client=mock_client
    )

    # Verify ThinkingSession was created correctly
    mock_thinking_session_class.assert_called_once_with(client=mock_client, session_id="test-session-id")

    # Verify record_thought was called for each thought
    assert mock_session.record_thought.call_count == 3

    # Check first thought call
    first_call = mock_session.record_thought.call_args_list[0][1]
    assert first_call["thought"] == "Thought 1"
    assert first_call["thought_number"] == 1
    assert first_call["total_thoughts"] == 3
    assert first_call["next_thought_needed"] is True

    # Check last thought call
    last_call = mock_session.record_thought.call_args_list[2][1]
    assert last_call["thought"] == "Thought 3"
    assert last_call["thought_number"] == 3
    assert last_call["total_thoughts"] == 3
    assert last_call["next_thought_needed"] is False

    # Verify result contains expected keys
    assert "session_id" in result
    assert "total_thoughts" in result
    assert "metadata" in result
    assert "summary" in result
    assert result["session_id"] == "test-session-id"
    assert result["total_thoughts"] == 3
    assert result["metadata"] == metadata


def test_find_thoughts_across_sessions(mock_client):
    """Test finding thoughts across sessions."""
    results = find_thoughts_across_sessions(
        query="test query",
        n_results=5,
        threshold=0.7,
        include_branches=False,
        session_id="specific-session-id",
        client=mock_client,
    )

    # Verify client method was called correctly
    mock_client.mcp_chroma_dev_chroma_find_similar_thoughts.assert_called_once()
    call_args = mock_client.mcp_chroma_dev_chroma_find_similar_thoughts.call_args[1]
    assert call_args["query"] == "test query"
    assert call_args["n_results"] == 5
    assert call_args["threshold"] == 0.7
    assert call_args["include_branches"] is False
    assert call_args["session_id"] == "specific-session-id"

    # Verify results
    assert len(results) == 1
    assert results[0]["document"] == "Test thought"
    assert results[0]["metadata"]["session_id"] == "test-session-id"


@patch("chroma_mcp_thinking.utils.ThinkingSession")
def test_create_thought_branch(mock_thinking_session_class, mock_client):
    """Test creating a thought branch."""
    # Setup mock session
    mock_session = MagicMock()
    mock_session.session_id = "parent-session-id"
    mock_session.get_session_summary.return_value = {"thoughts": [{"document": "Branch thought"}]}
    mock_thinking_session_class.return_value = mock_session

    # Call the function
    branch_thoughts = ["Branch thought 1", "Branch thought 2"]
    branch_id = "test-branch"
    result = create_thought_branch(
        parent_session_id="parent-session-id",
        parent_thought_number=2,
        branch_thoughts=branch_thoughts,
        branch_id=branch_id,
        client=mock_client,
    )

    # Verify ThinkingSession was created correctly
    mock_thinking_session_class.assert_called_once_with(client=mock_client, session_id="parent-session-id")

    # Verify record_thought was called for each thought
    assert mock_session.record_thought.call_count == 2

    # Check first thought call (should include branch_from_thought)
    first_call = mock_session.record_thought.call_args_list[0][1]
    assert first_call["thought"] == "Branch thought 1"
    assert first_call["thought_number"] == 1
    assert first_call["total_thoughts"] == 2
    assert first_call["branch_id"] == "test-branch"
    assert first_call["branch_from_thought"] == 2
    assert first_call["next_thought_needed"] is True

    # Check second thought call
    second_call = mock_session.record_thought.call_args_list[1][1]
    assert second_call["thought"] == "Branch thought 2"
    assert second_call["thought_number"] == 2
    assert second_call["total_thoughts"] == 2
    assert second_call["branch_id"] == "test-branch"
    assert second_call["branch_from_thought"] == 0  # Only the first thought links to parent
    assert second_call["next_thought_needed"] is False

    # Verify result contains expected keys
    assert "session_id" in result
    assert "branch_id" in result
    assert "parent_thought" in result
    assert "total_branch_thoughts" in result
    assert "summary" in result
    assert result["session_id"] == "parent-session-id"
    assert result["branch_id"] == "test-branch"
    assert result["parent_thought"] == 2
    assert result["total_branch_thoughts"] == 2


def test_create_thought_branch_invalid_parent():
    """Test creating a thought branch with invalid parent thought number."""
    with pytest.raises(ValueError):
        create_thought_branch(
            parent_session_id="parent-session-id",
            parent_thought_number=0,  # Invalid, must be at least 1
            branch_thoughts=["Branch thought"],
        )



"""Test configuration and fixtures for ChromaDB MCP server tests."""

import pytest
import chromadb
import inspect
import uuid
import time
import logging
import logging.handlers
import os
import warnings
import argparse
import sys

# Import the server module to access its globals
from src.chroma_mcp import server

from typing import Dict, List, Optional, Any
from datetime import datetime

from unittest.mock import AsyncMock, MagicMock, patch
from chromadb.types import Collection

from mcp.shared.exceptions import McpError
from mcp.types import ErrorData, INTERNAL_ERROR, INVALID_PARAMS

from src.chroma_mcp.types import ChromaClientConfig, ThoughtMetadata
from src.chroma_mcp.utils import (
    get_chroma_client,
    get_embedding_function,
    ValidationError,
    set_main_logger,
    set_server_config,
    BASE_LOGGER_NAME,
)

from unittest.mock import MagicMock, patch
from dotenv import load_dotenv

# Import necessary components for testing server configuration and setup
from src.chroma_mcp.server import config_server

# Import functions/classes needed for tool tests
from src.chroma_mcp.utils.chroma_client import get_chroma_client, get_embedding_function
from src.chroma_mcp.utils.errors import McpError, ValidationError
from src.chroma_mcp.types import (
    ChromaClientConfig,
    ThoughtMetadata,
    DocumentMetadata,  # Ensure DocumentMetadata is imported
)

# Import the function to initialize the client
from chroma_mcp.server import _initialize_chroma_client

# Import the reset function with its new name
from chroma_mcp.utils.chroma_client import reset_client as reset_chroma_client

# --- Start: Logger Configuration for Tests ---
TEST_LOG_DIR = "logs"
# Target the actual base logger name used by the application
# TEST_BASE_LOGGER_NAME = "chromamcp.test" # Use a sub-logger for tests
TEST_BASE_LOGGER_NAME = "chromamcp"  # Configure the app's root logger

logger = logging.getLogger(TEST_BASE_LOGGER_NAME)
logger.setLevel(logging.DEBUG)  # Default to DEBUG for tests

# Prevent adding handlers multiple times
if not logger.hasHandlers():
    formatter = logging.Formatter(
        f"%(asctime)s | %(name)-{len(TEST_BASE_LOGGER_NAME)+10}s | %(levelname)-8s | %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    # Console handler (useful for seeing test output)
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    # File handler
    os.makedirs(TEST_LOG_DIR, exist_ok=True)
    log_file = os.path.join(TEST_LOG_DIR, "test_debug.log")
    file_handler = logging.handlers.RotatingFileHandler(log_file, maxBytes=5 * 1024 * 1024, backupCount=3)  # 5 MB
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

# --- Explicitly set the global instance for tests ---
server._main_logger_instance = logger
# --- End: Logger Configuration for Tests ---

THOUGHTS_COLLECTION = "thoughts"
DEFAULT_SIMILARITY_THRESHOLD = 0.7


# Define MockCollection at module level
class MockCollection:
    """Mock implementation of ChromaDB collection."""

    def __init__(self):
        self.name = "test_collection"
        self.metadata = {"description": "test collection"}

    def add(self, documents=None, metadatas=None, ids=None, embeddings=None):
        """Mock add method."""
        return None

    def get(self, ids=None, where=None, limit=None, offset=None, where_document=None):
        """Mock get method."""
        return {
            "ids": ["1", "2"],
            "documents": ["thought1", "thought2"],
            "metadatas": [
                {"session_id": "test_session", "thought_number": 1},
                {"session_id": "test_session", "thought_number": 2},
            ],
            "embeddings": None,
        }

    def query(self, query_texts=None, n_results=2, where=None, where_document=None):
        """Mock query method."""
        return {
            "ids": [["1", "2"]],
            "documents": [["thought1", "thought2"]],
            "metadatas": [
                [
                    {"session_id": "test_session", "thought_number": 1},
                    {"session_id": "test_session", "thought_number": 2},
                ]
            ],
            "distances": [[0.1, 0.2]],
        }

    def count(self):
        """Mock count method."""
        return 2

    def modify(self, *args, **kwargs):
        """Mock modify method."""
        return None

    def delete(self, *args, **kwargs):
        """Mock delete method."""
        return None

    def update(self, *args, **kwargs):
        """Mock update method."""
        return None

    def upsert(self, *args, **kwargs):
        """Mock upsert method."""
        return None

    def peek(self, limit=10):
        """Mock peek method."""
        return self.get(limit=limit)

    def __str__(self):
        return f"MockCollection(name={self.name})"


# Define MockClient at module level
class MockClient:
    """Mock implementation of ChromaDB client."""

    def get_collection(self, *args, **kwargs):
        return MockCollection()

    def list_collections(self):
        return []


@pytest.fixture
def mock_chroma_client():
    """Mock ChromaDB client."""
    return MockClient()


@pytest.fixture
def mock_collection():
    """Mock ChromaDB collection."""
    return MockCollection()


@pytest.fixture
def mock_config() -> ChromaClientConfig:
    """Create a mock ChromaDB client configuration."""
    return ChromaClientConfig(client_type="ephemeral", host=None, port=None, data_dir=None)


@pytest.fixture
def sample_documents():
    """Create sample documents for testing."""
    return {
        "documents": ["doc1", "doc2"],
        "metadatas": [{"key": "value1"}, {"key": "value2"}],
        "ids": ["1", "2"],
        "embeddings": [[0.1, 0.2], [0.3, 0.4]],
    }


@pytest.fixture
def sample_thought() -> ThoughtMetadata:
    """Create a sample thought metadata."""
    return ThoughtMetadata(
        session_id="test_session",
        thought_number=1,
        total_thoughts=3,
        timestamp=1234567890,
        branch_from_thought=None,
        branch_id=None,
        next_thought_needed=False,
        custom_data={"key": "value"},
    )


@pytest.fixture
def sample_session():
    """Create a sample thinking session for testing."""
    return {
        "session_id": "test_session",
        "metadata": {"total_thoughts": 3, "start_time": 1234567890, "status": "in_progress"},
        "thoughts": [{"thought": "Initial thought", "metadata": {"thought_number": 1, "total_thoughts": 3}}],
    }


class MockMCP:
    """Mock MCP class with all required methods."""

    def __init__(self):
        """Initialize mock MCP."""
        self.name = "mock-mcp"

    async def chroma_add_documents(
        self,
        collection_name: str,
        documents: List[str],
        metadatas: Optional[List[Dict[str, Any]]] = None,
        ids: Optional[List[str]] = None,
    ) -> Dict[str, Any]:
        """Add documents to a collection."""
        # Validate inputs
        if not documents:
            raise McpError(ErrorData(code=INVALID_PARAMS, message="No documents provided"))

        return {"status": "success", "collection_name": collection_name, "count": len(documents)}

    async def chroma_query_documents(
        self,
        collection_name: str,
        query_texts: List[str],
        n_results: int = 10,
        where: Optional[Dict[str, Any]] = None,
        where_document: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Query documents from a collection."""
        # Generate mock results
        results = []
        for query in query_texts:
            matches = []
            for i in range(n_results):
                matches.append(
                    {
                        "id": f"{i+1}",
                        "document": f"doc{i+1}",
                        "metadata": {"key": f"value{i+1}"},
                        "distance": 0.1 * (i + 1),
                    }
                )
            results.append({"query": query, "matches": matches})

        return {"results": results}

    async def chroma_get_documents(
        self,
        collection_name: str,
        ids: Optional[List[str]] = None,
        where: Optional[Dict[str, Any]] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Dict[str, Any]:
        """Get documents from a collection."""
        # Generate mock results
        documents = []
        if ids:
            for id in ids:
                documents.append({"id": id, "document": f"doc-{id}", "metadata": {"key": f"value-{id}"}})
        else:
            for i in range(1, (limit or 10) + 1):
                documents.append({"id": f"{i}", "document": f"doc{i}", "metadata": {"key": f"value{i}"}})

        return {"documents": documents}

    async def chroma_update_documents(
        self,
        collection_name: str,
        ids: List[str],
        documents: Optional[List[str]] = None,
        metadatas: Optional[List[Dict[str, Any]]] = None,
    ) -> Dict[str, Any]:
        """Update documents in a collection."""
        return {"status": "success", "collection_name": collection_name, "count": len(ids)}

    async def chroma_delete_documents(
        self, collection_name: str, ids: Optional[List[str]] = None, where: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """Delete documents from a collection."""
        count = len(ids) if ids else 1
        return {"status": "success", "collection_name": collection_name, "count": count}

    async def chroma_create_collection(
        self,
        collection_name: str,
        description: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        hnsw_space: Optional[str] = None,
        hnsw_construction_ef: Optional[int] = None,
        hnsw_search_ef: Optional[int] = None,
        hnsw_M: Optional[int] = None,
    ) -> Dict[str, Any]:
        """Create a new collection."""
        # Create collection metadata
        collection_metadata = {
            "description": description or f"Collection {collection_name}",
            "settings": {
                "hnsw:space": hnsw_space or "l2",
                "hnsw:construction_ef": hnsw_construction_ef or 100,
                "hnsw:search_ef": hnsw_search_ef or 100,
                "hnsw:M": hnsw_M or 16,
            },
        }

        if metadata:
            collection_metadata.update(metadata)

        return {"name": collection_name, "id": str(uuid.uuid4()), "metadata": collection_metadata}

    async def chroma_list_collections(
        self, limit: Optional[int] = None, offset: Optional[int] = None, name_contains: Optional[str] = None
    ) -> Dict[str, Any]:
        """List available collections."""
        # Generate mock collections
        collections = []
        for i in range(1, 4):
            name = f"collection{i}"
            if name_contains and name_contains not in name:
                continue
            collections.append({"name": name, "id": f"{i}", "metadata": {"description": f"Description for {name}"}})

        # Apply limit and offset
        if offset:
            collections = collections[offset:]
        if limit:
            collections = collections[:limit]

        return {"collections": collections, "total_count": len(collections)}

    async def chroma_get_collection(self, collection_name: str) -> Dict[str, Any]:
        """Get information about a collection."""
        return {
            "name": collection_name,
            "id": str(uuid.uuid4()),
            "metadata": {"description": f"Description for {collection_name}"},
            "count": 10,
            "sample_entries": [{"id": "1", "document": "Sample doc 1"}, {"id": "2", "document": "Sample doc 2"}],
        }

    async def chroma_modify_collection(
        self, collection_name: str, new_metadata: Optional[Dict[str, Any]] = None, new_name: Optional[str] = None
    ) -> Dict[str, Any]:
        """Modify an existing collection."""
        modified_name = new_name or collection_name
        modified_metadata = {"description": f"Description for {modified_name}"}
        if new_metadata:
            modified_metadata.update(new_metadata)

        return {"name": modified_name, "id": str(uuid.uuid4()), "metadata": modified_metadata}

    async def chroma_delete_collection(self, collection_name: str) -> Dict[str, Any]:
        """Delete a collection."""
        return {"status": "success", "collection_name": collection_name}

    async def chroma_peek_collection(self, collection_name: str, limit: int = 10) -> Dict[str, Any]:
        """Peek at documents in a collection."""
        # Generate mock entries
        entries = []
        for i in range(1, limit + 1):
            entries.append({"id": f"{i}", "document": f"Document {i}", "metadata": {"key": f"value{i}"}})

        return {"entries": entries}

    async def chroma_sequential_thinking(
        self,
        thought: str,
        session_id: str,
        thought_number: int,
        total_thoughts: int,
        branch_id: Optional[str] = None,
        branch_from_thought: Optional[int] = None,
        next_thought_needed: bool = False,
        custom_data: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Process sequential thoughts."""
        # Validate inputs
        if not thought:
            raise McpError(ErrorData(code=INVALID_PARAMS, message="Thought content is required"))
        if thought_number < 1:
            raise McpError(ErrorData(code=INVALID_PARAMS, message=f"Invalid thought number: {thought_number}"))

        response = {
            "status": "success",
            "thought": thought,
            "session_id": session_id,
            "thought_number": thought_number,
            "total_thoughts": total_thoughts,
        }

        # Add branch information if provided
        if branch_id:
            response["branch_id"] = branch_id
        if branch_from_thought:
            response["branch_from_thought"] = branch_from_thought

        # Add previous thought if it exists
        if thought_number > 1:
            response["previous_thought"] = "previous thought"

        return response

    async def chroma_find_similar_thoughts(
        self,
        query: str,
        n_results: int = 5,
        session_id: Optional[str] = None,
        threshold: float = DEFAULT_SIMILARITY_THRESHOLD,
        include_branches: bool = True,
    ) -> Dict[str, Any]:
        """Find similar thoughts."""
        matches = [
            {"thought": "thought1", "metadata": {"session_id": "session1"}, "similarity": 0.9},
            {"thought": "thought2", "metadata": {"session_id": "session2"}, "similarity": 0.8},
        ]

        # Filter by session if specified
        if session_id:
            matches = [m for m in matches if m["metadata"]["session_id"] == session_id]

        return {"matches": matches[:n_results]}

    async def chroma_get_session_summary(self, session_id: str, include_branches: bool = True) -> Dict[str, Any]:
        """Get summary for a session."""
        return {
            "session_id": session_id,
            "thoughts": [{"thought": "thought1", "thought_number": 1}, {"thought": "thought2", "thought_number": 2}],
        }

    async def chroma_find_similar_sessions(
        self, query: str, n_results: int = 3, threshold: float = DEFAULT_SIMILARITY_THRESHOLD
    ) -> Dict[str, Any]:
        """Find similar sessions."""
        return {
            "matches": [
                {"session_id": "session1", "summary": "session1 summary", "similarity": 0.9},
                {"session_id": "session2", "summary": "session2 summary", "similarity": 0.8},
            ][:n_results]
        }

    async def ping(self) -> Dict[str, Any]:
        """Test if the MCP server is alive."""
        return {"status": "ok", "message": "Server is alive"}


@pytest.fixture(scope="session")
def patched_mcp():
    """Return a mock MCP instance with all required methods."""
    return MockMCP()


# Add this fixture to automatically shut down logging after tests
@pytest.fixture(autouse=True)
def shutdown_logging_after_tests():
    """Ensures logging is shut down after tests, closing handlers."""
    yield
    logging.shutdown()


@pytest.fixture(scope="function")
def initialized_chroma_client():
    """Fixture to ensure chroma client is initialized before a test runs.

    Uses ephemeral client by default. Resets after test.
    """
    # Create mock args for initialization (using ephemeral client)
    mock_args = argparse.Namespace(
        dotenv_path=None,  # Assume no .env for basic test
        cpu_execution_provider="auto",
        client_type="ephemeral",
        data_dir=None,
        host=None,
        port=8000,
        ssl=False,
        tenant=None,
        database=None,
        api_key=None,
        embedding_function_name="default",
        # Add other args _initialize_chroma_client might access, even via getattr
        # Though getattr should handle missing ones
    )

    # Patch get_logger to avoid side effects during init
    with patch("chroma_mcp.server.get_logger") as mock_get_logger:
        # Ensure chromadb is available or mocked if needed for specific tests
        # For basic initialization testing, assuming it's available
        # Call the actual initialization function
        _initialize_chroma_client(mock_args)

    yield  # Test runs here

    # Teardown: Reset the client instance after the test
    # This assumes you have a way to reset the global _chroma_client_instance
    # Option 1: If you have a reset function (best)
    reset_chroma_client()
    # Option 2: Manually reset the global (more brittle)
    # from chroma_mcp import server
    # server._chroma_client_instance = None
    # print("Chroma client reset after test.")


# --- Begin added logger fixture ---
@pytest.fixture(scope="session", autouse=True)
def configure_test_logger():
    """
    Configures a basic logger for the test session to prevent
    'Logger requested before main configuration' warnings and ensure
    logs are captured by pytest if needed. Runs automatically for the session.
    """
    logger = logging.getLogger(BASE_LOGGER_NAME)
    # Prevent adding handlers multiple times if fixture runs unexpectedly often
    if not logger.hasHandlers():
        logger.setLevel(logging.DEBUG)  # Set to DEBUG to capture all levels during tests
        handler = logging.StreamHandler(sys.stderr)  # Log to stderr like pytest does
        formatter = logging.Formatter("%(asctime)s | %(name)-20s | %(levelname)-8s | %(message)s")
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        # Set this basic logger as the 'main' logger for the utils.get_logger() function
        set_main_logger(logger)
        logger.debug("Test logger configured by session fixture.")
    # Yield control to the test session
    yield
    # Teardown (optional, e.g., remove handler if necessary)
    # logger.handlers.clear() # Might not be needed if process exits cleanly


# --- End added logger fixture ---


# Fixture to initialize Chroma client once per session for specific tests
# This avoids repeated initialization warnings if multiple tests need it.
# Note: This assumes a persistent client setup for testing to avoid state issues.
# If ephemeral is needed, adjust or use a function-scoped fixture.
@pytest.fixture(scope="session")
def initialized_chroma_client(tmp_path_factory):
    """Initializes a persistent Chroma client for the test session."""
    data_path = tmp_path_factory.mktemp("chroma_test_data")
    print(f"DEBUG: Using tmp Chroma data path: {data_path}")

    # Define a minimal config for testing
    test_config = ChromaClientConfig(
        client_type="persistent",
        data_dir=str(data_path),
        embedding_function_name="default",  # Use a known default EF
    )

    # Store the original config if it exists, to restore later
    # Note: get_server_config might raise if never set, handle this?
    # For simplicity, assume it's None initially or we don't need strict restore.
    # original_config = None
    # try:
    #     original_config = get_server_config()
    # except Exception:
    #     pass

    # Set the test configuration globally
    set_server_config(test_config)
    print(f"DEBUG: Set test server config: {test_config}")

    # Use patch only for the client instance, ensuring it starts as None
    with patch("src.chroma_mcp.utils.chroma_client._chroma_client", None):
        print("DEBUG: Attempting to get Chroma client within fixture...")
        try:
            # Call get_chroma_client which should initialize based on the *globally set* config
            client = get_chroma_client()
            print(f"DEBUG: Chroma client obtained: {client}")
            yield client  # Provide the client to tests

            # Teardown:
            # Reset the client instance via the utility function
            from src.chroma_mcp.utils.chroma_client import reset_client

            reset_client()
            print("DEBUG: Chroma client reset via reset_client().")

            # Restore original config if needed (optional)
            # if original_config:
            #    set_server_config(original_config)
            # else:
            #    # Or clear it? Depends on desired state after tests.
            #    # You might need a `clear_server_config()` function.
            #    pass
            print("DEBUG: Chroma client fixture teardown complete.")

        except Exception as e:
            print(f"ERROR: Failed to initialize/yield Chroma client in fixture: {e}")
            pytest.fail(f"Failed to initialize Chroma client for tests: {e}")



"""
Command Line Interface for the Direct ChromaDB Client.

Provides commands for indexing, querying, and managing the ChromaDB instance
used for automation tasks like RAG context building.
"""

import argparse
import os
import sys
import logging
from pathlib import Path
from dotenv import load_dotenv
import uuid
import time

# Get our specific logger
logger = logging.getLogger(__name__)

# Removed sys.path manipulation logic
# Imports should work directly when the package is installed
from .connection import get_client_and_ef
from .indexing import index_file, index_git_files
from .analysis import analyze_chat_history  # Import the new function

# Import from utils to set the main logger
from chroma_mcp.utils import set_main_logger, BASE_LOGGER_NAME as UTILS_BASE_LOGGER_NAME

# Import the function for the new subcommand
from .interactive_promoter import run_interactive_promotion

# Import the refactored promotion function
from .learnings import promote_to_learnings_collection

# --- Constants ---
DEFAULT_COLLECTION_NAME = "codebase_v1"
DEFAULT_QUERY_RESULTS = 5

# Load environment variables from .env file
load_dotenv()

# Determine default log level from environment or fallback
default_log_level_env = os.getenv("LOG_LEVEL", "INFO").upper()
if default_log_level_env not in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
    default_log_level_env = "INFO"  # Fallback if invalid value in env


def main():
    """Main entry point for the chroma-client CLI."""
    # Configure logging EARLY inside main()
    log_format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    # Set a default level; will be adjusted based on verbosity later
    logging.basicConfig(level=logging.WARNING, format=log_format)

    # Get the logger for the current module (chroma_mcp_client.cli)
    # This specific logger is fine, but we also need to set the main one for utils.
    logger = logging.getLogger(__name__)

    # --- Setup Logging Level based on verbosity (occurs after arg parsing) ---
    # The actual setting of levels will happen after args are parsed.
    # For now, we ensure a basic config is up.

    parser = argparse.ArgumentParser(
        description="ChromaDB Client CLI for indexing and querying.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase output verbosity (e.g., -v for INFO, -vv for DEBUG)",
    )

    subparsers = parser.add_subparsers(dest="command", required=True, help="Available commands")

    # --- Index Subparser ---
    index_parser = subparsers.add_parser(
        "index",
        help="Index specific files or all git-tracked files into ChromaDB.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    index_parser.add_argument(
        "paths",
        nargs="*",
        help="Optional specific file or directory paths to index.",
        type=Path,
        default=[],
    )
    index_parser.add_argument(
        "--repo-root",
        type=Path,
        default=Path(os.getcwd()),
        help="Repository root path (used for determining relative paths for IDs).",
    )
    index_parser.add_argument(
        "--all", action="store_true", help="Index all files currently tracked by git in the repo."
    )
    index_parser.add_argument(
        "--collection-name",
        default=DEFAULT_COLLECTION_NAME,
        help="Name of the ChromaDB collection to use.",
    )

    # --- Count Subparser ---
    count_parser = subparsers.add_parser("count", help="Count documents in a ChromaDB collection.")
    count_parser.add_argument(
        "--collection-name",
        default=DEFAULT_COLLECTION_NAME,
        help="Name of the ChromaDB collection to count.",
    )

    # --- Query Subparser ---
    query_parser = subparsers.add_parser("query", help="Query a ChromaDB collection.")
    query_parser.add_argument("query_text", help="The text to query for.")
    query_parser.add_argument(
        "--collection-name",
        default=DEFAULT_COLLECTION_NAME,
        help="Name of the ChromaDB collection to query.",
    )
    query_parser.add_argument(
        "-n",
        "--n-results",
        type=int,
        default=DEFAULT_QUERY_RESULTS,
        help="Number of results to return.",
    )

    # --- Analyze Chat History Subparser ---
    analyze_parser = subparsers.add_parser(
        "analyze-chat-history",
        help="Analyze recent chat history to correlate with Git changes.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    analyze_parser.add_argument(
        "--collection-name",
        default="chat_history_v1",
        help="Name of the ChromaDB chat history collection.",
    )
    analyze_parser.add_argument(
        "--repo-path",
        type=Path,
        default=Path(os.getcwd()),
        help="Path to the Git repository to analyze.",
    )
    analyze_parser.add_argument(
        "--status-filter",
        default="captured",
        help="Metadata status value to filter entries for analysis.",
    )
    analyze_parser.add_argument(
        "--new-status",
        default="analyzed",
        help="Metadata status value to set after analysis.",
    )
    analyze_parser.add_argument(
        "--days-limit",
        type=int,
        default=7,
        help="How many days back to look for entries to analyze.",
    )

    # --- Update Collection EF Subparser ---
    update_ef_parser = subparsers.add_parser(
        "update-collection-ef",
        help="Update the embedding function name stored in a collection's metadata.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    update_ef_parser.add_argument(
        "--collection-name",
        required=True,
        help="Name of the ChromaDB collection to update.",
    )
    update_ef_parser.add_argument(
        "--ef-name",
        required=True,
        help="The new embedding function name string to store (e.g., 'sentence_transformer').",
    )

    # --- Setup Collections Subparser ---
    setup_collections_parser = subparsers.add_parser(
        "setup-collections",
        help="Check and create all required ChromaDB collections if they do not exist.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    # No arguments needed for this command yet, but parser is set up for future extension

    # --- Review and Promote Subparser ---
    review_promote_parser = subparsers.add_parser(
        "review-and-promote",
        help="Interactively review 'analyzed' chat entries and promote them to learnings.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    review_promote_parser.add_argument(
        "--days-limit",
        type=int,
        default=7,
        help="How many days back to look for 'analyzed' entries.",
    )
    review_promote_parser.add_argument(
        "--fetch-limit",
        type=int,
        default=50,
        help="Maximum number of entries to fetch for review in one go.",
    )
    review_promote_parser.add_argument(
        "--chat-collection-name",
        default="chat_history_v1",
        help="Name of the ChromaDB chat history collection.",
    )
    review_promote_parser.add_argument(
        "--learnings-collection-name",
        default="derived_learnings_v1",
        help="Name of the collection to add derived learnings to.",
    )
    # Add --repo-path later if needed for diffs

    # --- Promote Learning Subparser ---
    promote_parser = subparsers.add_parser(
        "promote-learning",
        help="Promote an analyzed insight or manual finding into the derived learnings collection.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    promote_parser.add_argument(
        "--description",
        required=True,
        help="Natural language description of the learning (will be embedded).",
    )
    promote_parser.add_argument(
        "--pattern",
        required=True,
        help="Core pattern identified (e.g., code snippet, regex, textual description).",
    )
    promote_parser.add_argument(
        "--code-ref",
        required=True,
        help="Code reference illustrating the learning (e.g., chunk_id 'path:sha:index').",
    )
    promote_parser.add_argument(
        "--tags",
        required=True,
        help="Comma-separated tags for categorization (e.g., 'python,refactor,logging').",
    )
    promote_parser.add_argument(
        "--confidence",
        required=True,
        type=float,
        help="Confidence score for this learning (e.g., 0.0 to 1.0).",
    )
    promote_parser.add_argument(
        "--source-chat-id",
        help="Optional ID of the source entry in the chat history collection.",
    )
    promote_parser.add_argument(
        "--collection-name",
        default="derived_learnings_v1",
        help="Name of the collection to add the derived learning to.",
    )
    promote_parser.add_argument(
        "--chat-collection-name",
        default="chat_history_v1",
        help="Name of the chat history collection to update status if source ID is provided.",
    )

    args = parser.parse_args()

    # --- Setup Logging Level based on verbosity ---
    if args.verbose == 1:
        log_level = logging.INFO
    elif args.verbose >= 2:
        log_level = logging.DEBUG
    else:
        # Default to INFO if not specified or 0, to ensure our logs are seen
        # unless the user *really* wants less (which isn't an option here)
        log_level = logging.INFO

    # Set the level for the root logger AND the main 'chromamcp' base logger from utils
    # This affects all loggers unless they have their own level set
    logging.getLogger().setLevel(log_level)
    logging.getLogger(UTILS_BASE_LOGGER_NAME).setLevel(log_level)  # Configure the base 'chromamcp' logger

    # Set the main logger instance for chroma_mcp.utils
    # This uses the already configured 'chromamcp' base logger.
    set_main_logger(logging.getLogger(UTILS_BASE_LOGGER_NAME))

    # Set the level specifically for our client's own sub-loggers if desired
    # (e.g., chroma_mcp_client.cli, chroma_mcp_client.connection)
    # This ensures that even if the root/base is INFO, client can be DEBUG if -vv
    logging.getLogger("chroma_mcp_client").setLevel(log_level)

    # Optionally set levels for other loggers if needed (e.g., sentence_transformers)
    logging.getLogger("sentence_transformers").setLevel(logging.WARNING)

    # logger.info(f"Log level set to {logging.getLevelName(log_level)}")
    # Use the utils logger to announce this, as it's now configured.
    utils_logger = logging.getLogger(f"{UTILS_BASE_LOGGER_NAME}.cli_setup")  # Get a child of the main logger
    utils_logger.info(
        f"Client CLI log level set to {logging.getLevelName(log_level)} for base '{UTILS_BASE_LOGGER_NAME}' and 'chroma_mcp_client'"
    )

    # Initialize client/EF early to catch connection errors
    try:
        logger.info("Initializing ChromaDB connection...")
        client, ef = get_client_and_ef()  # Loads config from .env via connection module
        logger.info(
            f"ChromaDB connection successful (Client type: {client.__class__.__name__}, EF: {ef.__class__.__name__ if ef else 'Default'})"
        )
    except Exception as e:
        logger.critical(f"Failed to initialize ChromaDB client: {e}", exc_info=True)
        sys.exit(1)

    # --- Command Handling ---
    if args.command == "index":
        collection_name = args.collection_name
        logger.info(f"Executing 'index' command for collection '{collection_name}'...")
        # Resolve repo root path - specific to index command
        repo_root_path = args.repo_root.resolve()
        # REMOVED: Premature get_or_create_collection call
        # collection = client.get_or_create_collection(
        #     name=collection_name,
        #     embedding_function=ef,  # Pass embedding function here
        #     # Add metadata if needed: metadata=get_collection_settings(...)
        # )

        if args.all:
            logger.info(f"Indexing all tracked git files in {repo_root_path}...")
            # Pass the collection_name string
            count = index_git_files(repo_root_path, collection_name)
            logger.info(f"Git index command finished. Indexed {count} files.")
        elif args.paths:
            logger.info(f"Processing {len(args.paths)} specified file/directory paths...")
            indexed_count = 0
            for path_item in args.paths:
                if path_item.is_file():
                    # Pass the collection_name string
                    if index_file(path_item, repo_root_path, collection_name):
                        indexed_count += 1
                elif path_item.is_dir():
                    logger.warning(
                        f"Skipping directory: {path_item}. Indexing directories directly is not yet supported."
                    )
                    # TODO: Implement recursive directory indexing if needed
                else:
                    logger.warning(f"Skipping non-existent path: {path_item}")
            logger.info(f"File/directory index command finished. Indexed {indexed_count} files.")
        else:
            logger.warning("Index command called without --all flag or specific paths. Nothing to index.")

    elif args.command == "count":
        collection_name = args.collection_name
        logger.info(f"Executing 'count' command for collection '{collection_name}'...")
        try:
            collection = client.get_collection(name=collection_name)
            count = collection.count()
            print(f"Collection '{collection_name}' contains {count} documents.")
        except Exception as e:
            logger.error(f"Failed to get count for collection '{collection_name}': {e}", exc_info=True)
            print(f"Error: Could not retrieve collection '{collection_name}'. Does it exist?", file=sys.stderr)
            sys.exit(1)

    elif args.command == "query":
        collection_name = args.collection_name
        logger.info(f"Executing 'query' command for collection '{collection_name}'...")
        logger.info(f"Query text: '{args.query_text}'")
        logger.info(f"Number of results: {args.n_results}")
        try:
            collection = client.get_collection(name=collection_name, embedding_function=ef)
            results = collection.query(
                query_texts=[args.query_text], n_results=args.n_results, include=["metadatas", "documents", "distances"]
            )

            print(f"\n--- Query Results for '{args.query_text}' ({args.n_results} requested) ---")
            if not results or not results.get("ids") or not results["ids"][0]:
                print("No results found.")
            else:
                # Results is a dict with lists for each query_text (we only have one)
                ids = results["ids"][0]
                distances = results["distances"][0]
                metadatas = results["metadatas"][0]
                documents = results["documents"][0]

                for i, doc_id in enumerate(ids):
                    print(f"\nResult {i+1}:")
                    print(f"  ID: {doc_id}")
                    print(f"  Distance: {distances[i]:.4f}")
                    if metadatas and metadatas[i]:
                        print(f"  Metadata: {metadatas[i]}")
                    if documents and documents[i]:
                        # Show first N characters as snippet
                        snippet = documents[i].replace("\n", " ").strip()
                        max_snippet_len = 150
                        print(
                            f"  Snippet: {snippet[:max_snippet_len]}{'...' if len(snippet) > max_snippet_len else ''}"
                        )
                print("---------------------------------------------")

        except Exception as e:
            logger.error(f"Failed to query collection '{collection_name}': {e}", exc_info=True)
            print(f"Error: Could not query collection '{collection_name}'. Does it exist?", file=sys.stderr)
            sys.exit(1)

    elif args.command == "analyze-chat-history":
        logger.info(f"Executing 'analyze-chat-history' command...")
        # Note: The analysis function uses the client and ef initialized earlier
        try:
            processed, correlated = analyze_chat_history(
                client=client,  # Pass the initialized client
                embedding_function=ef,  # Pass the initialized EF
                repo_path=str(args.repo_path.resolve()),  # Pass resolved absolute path
                collection_name=args.collection_name,
                days_limit=args.days_limit,
                # Pass limit from args if needed, or keep default from analysis function
                # limit=args.limit, # Assuming limit arg exists or is added
                status_filter=args.status_filter,
                new_status=args.new_status,
            )
            logger.info("'analyze-chat-history' command finished. Processed=%d, Correlated=%d", processed, correlated)
            print("Chat history analysis finished.")  # Keep simple user output
        except Exception as e:
            logger.error(f"An error occurred during chat history analysis: {e}", exc_info=True)
            print(f"Error during analysis: {e}")  # Inform user
            sys.exit(1)  # Exit with non-zero code on error

    elif args.command == "update-collection-ef":
        collection_name = args.collection_name
        ef_name = args.ef_name
        logger.info(
            f"Executing 'update-collection-ef' command for collection '{collection_name}' with new EF name '{ef_name}'..."
        )
        try:
            collection = client.get_collection(name=collection_name)  # Get it first to ensure it exists
            metadata = collection.metadata or {}
            metadata["embedding_function_name"] = ef_name
            collection.modify(metadata=metadata)
            logger.info(f"Successfully updated embedding function name for collection '{collection_name}'.")
            print(f"Embedding function name for '{collection_name}' updated to '{ef_name}'.")
        except Exception as e:
            logger.error(
                f"Failed to update embedding function name for collection '{collection_name}': {e}",
                exc_info=True,
            )
            print(
                f"Error: Could not update collection '{collection_name}'. Does it exist?",
                file=sys.stderr,
            )
            sys.exit(1)

    elif args.command == "setup-collections":
        logger.info("Executing 'setup-collections' command...")
        required_collections = [
            "codebase_v1",
            "chat_history_v1",
            "derived_learnings_v1",
            "thinking_sessions_v1",
        ]
        collections_created = 0
        collections_existing = 0

        for collection_name in required_collections:
            logger.info(f"Checking/Creating collection: '{collection_name}'")
            try:
                # Attempt to get the collection first to see if it exists
                # This is a bit more verbose but gives clearer logging
                try:
                    client.get_collection(name=collection_name)
                    logger.info(f"Collection '{collection_name}' already exists.")
                    collections_existing += 1
                except Exception:  # Should be a more specific exception if Chroma client provides one for not found
                    logger.info(f"Collection '{collection_name}' not found, creating...")
                    client.get_or_create_collection(
                        name=collection_name,
                        embedding_function=ef,
                        # metadata={"description": f"Default collection for {collection_name}"} # Example metadata
                    )
                    logger.info(f"Collection '{collection_name}' created successfully.")
                    collections_created += 1
            except Exception as e:
                logger.error(
                    f"Failed to check/create collection '{collection_name}': {e}",
                    exc_info=True,
                )
                print(
                    f"Error processing collection '{collection_name}'. See logs for details.",
                    file=sys.stderr,
                )
                # Optionally, decide if we should exit or continue with other collections
                # sys.exit(1) # Uncomment to exit on first error

        summary_message = (
            f"Collections setup finished. Created: {collections_created}, Already Existed: {collections_existing}."
        )
        logger.info(summary_message)
        print(summary_message)

    elif args.command == "review-and-promote":
        logger.info("Executing 'review-and-promote' command...")
        try:
            run_interactive_promotion(
                days_limit=args.days_limit,
                fetch_limit=args.fetch_limit,
                chat_collection_name=args.chat_collection_name,
                learnings_collection_name=args.learnings_collection_name,
                # repo_path=args.repo_path # Add if/when repo_path arg is added
            )
            logger.info("'review-and-promote' command finished.")
            print("Interactive review and promotion process complete.")
        except Exception as e:
            logger.error(f"An error occurred during interactive promotion: {e}", exc_info=True)
            print(f"Error during interactive promotion: {e}")
            sys.exit(1)

    elif args.command == "promote-learning":
        logger.info(f"Executing 'promote-learning' command...")

        # Call the refactored function
        # Note: client and ef are already initialized earlier in main()
        learning_id = promote_to_learnings_collection(
            client=client,
            embedding_function=ef,
            description=args.description,
            pattern=args.pattern,
            code_ref=args.code_ref,
            tags=args.tags,
            confidence=args.confidence,
            learnings_collection_name=args.collection_name,  # from --collection-name
            source_chat_id=args.source_chat_id,
            chat_history_collection_name=args.chat_collection_name,
        )

        if learning_id:
            # The function already prints messages, so we might not need more here
            # or we can have a simpler success message.
            logger.info(f"'promote-learning' command completed for learning ID: {learning_id}")
        else:
            logger.error("'promote-learning' command failed. See previous logs for details.")
            sys.exit(1)  # Exit if the promotion function indicated failure by returning None

    else:
        logger.error(f"Unknown command: {args.command}")


if __name__ == "__main__":
    main()



"""Tests for collection management tools."""

import pytest
import uuid
import json

from typing import Dict, Any, List, Optional
from unittest.mock import patch, MagicMock, AsyncMock, ANY, call
from contextlib import contextmanager

from mcp import types
from mcp.shared.exceptions import McpError
from mcp.types import INVALID_PARAMS, INTERNAL_ERROR, ErrorData

# Import specific errors if needed, or rely on ValidationError/Exception
from src.chroma_mcp.utils.errors import ValidationError

# --- Add direct import for setting global config ---
from src.chroma_mcp import utils as chroma_utils

# --- Add import for reset_client ---
from src.chroma_mcp.utils import chroma_client as client_utils

from src.chroma_mcp.tools.collection_tools import (
    _reconstruct_metadata,  # Keep helper if used
    _create_collection_impl,
    _create_collection_with_metadata_impl,
    _list_collections_impl,
    _get_collection_impl,
    _rename_collection_impl,
    _delete_collection_impl,
    _peek_collection_impl,
)

# Import Pydantic models used by the tools
from src.chroma_mcp.tools.collection_tools import (
    CreateCollectionInput,
    CreateCollectionWithMetadataInput,
    ListCollectionsInput,
    GetCollectionInput,
    RenameCollectionInput,
    DeleteCollectionInput,
    PeekCollectionInput,
)

# Correct import for get_collection_settings
from src.chroma_mcp.utils.config import get_collection_settings
from src.chroma_mcp.types import ChromaClientConfig

# Add import for direct patching
from src.chroma_mcp.tools import collection_tools

DEFAULT_SIMILARITY_THRESHOLD = 0.7


# Mock embedding function classes used in chroma_client
class MockEmbeddingFunction:
    def __init__(self, name="default"):  # Add name for identification
        self.name = name

    def __call__(self, input: List[str]):
        # Simple mock implementation
        return [[0.1] * 10 for _ in input]  # Use 'input' here too


# --- Fixtures ---


@pytest.fixture(autouse=True)
def mock_logger():
    """Auto-mock the logger."""
    with patch("src.chroma_mcp.tools.collection_tools.get_logger") as mock_get_logger:
        mock_log_instance = MagicMock()
        mock_get_logger.return_value = mock_log_instance
        yield mock_log_instance


@pytest.fixture
def mock_chroma_client():
    """Mock the ChromaDB client instance."""
    mock_client = MagicMock()
    # Mock the create_collection method specifically
    mock_client.create_collection = MagicMock()
    # Mock get_collection, list_collections etc. as needed for other tests
    mock_client.get_collection = MagicMock()
    mock_client.list_collections = MagicMock(return_value=[])
    mock_client.delete_collection = MagicMock()

    with patch("chroma_mcp.tools.collection_tools.get_chroma_client", return_value=mock_client):
        yield mock_client


@pytest.fixture
def mock_embedding_functions():
    """Mock the get_embedding_function lookup."""
    mock_functions = {
        "default": MockEmbeddingFunction(name="default"),
        "fast": MockEmbeddingFunction(name="fast"),
        "accurate": MockEmbeddingFunction(name="accurate"),
        "openai": MockEmbeddingFunction(name="openai"),
    }

    def mock_lookup(name):
        func = mock_functions.get(name.lower())
        if func is None:
            raise McpError(ErrorData(code=INVALID_PARAMS, message=f"Unknown embedding function: {name}"))
        return func

    # Patch where get_embedding_function is looked up in the tools module
    with patch("src.chroma_mcp.tools.collection_tools.get_embedding_function", side_effect=mock_lookup) as mock_getter:
        yield mock_getter  # Yield the mock getter itself for assertions


@pytest.fixture
def mock_server_config():
    """Mock the server configuration retrieval."""
    # Make this instance unique per test potentially
    mock_config = ChromaClientConfig(client_type="ephemeral", embedding_function_name="default")
    original_config = chroma_utils._global_client_config
    chroma_utils._global_client_config = mock_config
    try:
        yield mock_config  # Yield the config object to modify it in tests
    finally:
        # Ensure global config is reset after test
        chroma_utils._global_client_config = original_config


@pytest.fixture
def mock_collection_settings():
    """Mock retrieval of default collection settings."""
    default_settings = {"hnsw:space": "cosine"}
    with patch("chroma_mcp.tools.collection_tools.get_collection_settings", return_value=default_settings):
        yield default_settings


@pytest.fixture
def mock_validate_name():
    """Mock collection name validation."""
    with patch("chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate:
        yield mock_validate


# --- Helper Functions ---


def assert_successful_json_result(
    result: List[types.TextContent],
    expected_data: Optional[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    """Asserts the tool result is a successful list containing valid JSON, returning the parsed data."""
    assert isinstance(result, list)
    assert len(result) > 0, "Result list cannot be empty for successful JSON result."
    content_item = result[0]
    assert isinstance(content_item, types.TextContent), f"Expected TextContent, got {type(content_item)}"
    assert content_item.type == "text", f"Expected content type 'text', got '{content_item.type}'"
    assert content_item.text is not None, "Text content cannot be None for JSON result."
    try:
        parsed_data = json.loads(content_item.text)
        assert isinstance(parsed_data, dict), f"Parsed JSON is not a dictionary, got {type(parsed_data)}"
    except (json.JSONDecodeError, AssertionError) as e:
        pytest.fail(f"Result content is not valid JSON: {e}\nContent: {content_item.text}")
    if expected_data is not None:
        assert parsed_data == expected_data, f"Parsed JSON data mismatch. Expected: {expected_data}, Got: {parsed_data}"
    return parsed_data


# Rework assert_error_result as a context manager for future use
@contextmanager
def assert_raises_mcp_error(expected_message: str):
    """Asserts that McpError is raised and optionally checks the error message."""
    try:
        yield
    except McpError as e:
        # Check both e.error_data (if exists) and e.args[0]
        message = ""
        if hasattr(e, "error_data") and hasattr(e.error_data, "message"):
            message = str(e.error_data.message)
        elif e.args and isinstance(e.args[0], ErrorData) and hasattr(e.args[0], "message"):
            message = str(e.args[0].message)
        else:
            message = str(e)  # Fallback to the exception string itself

        assert (
            expected_message in message
        ), f"Expected error message containing '{expected_message}' but got '{message}'"
        return


# Apply usefixtures to the whole class again
# Apply the new reset fixture as well
# Restore mock_server_config to class level
@pytest.mark.usefixtures("mock_embedding_functions", "mock_server_config")
class TestCollectionTools:
    """Test cases for collection management tools."""

    # --- _create_collection_impl Tests ---
    @pytest.mark.asyncio
    # REMOVE fixtures, use local patching
    async def test_create_collection_uses_default_ef_from_config(self):
        """Test that _create_collection_impl uses the EF name from server config."""
        collection_name = "test-default-ef"

        # --- Mocks --- #
        mock_collection = MagicMock()
        mock_collection.name = collection_name
        mock_collection.id = "uuid-123"
        default_settings = get_collection_settings()  # Get defaults
        stored_metadata = {f"chroma:setting:{k.replace(':', '_')}": v for k, v in default_settings.items()}
        mock_collection.metadata = stored_metadata
        mock_collection.count.return_value = 0

        mock_client_instance = MagicMock()
        mock_client_instance.create_collection.return_value = mock_collection

        mock_ef_instance = MockEmbeddingFunction(name="default")

        # --- Patching --- #
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate_name,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_chroma_client", return_value=mock_client_instance
            ) as mock_get_client,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_embedding_function", return_value=mock_ef_instance
            ) as mock_get_ef,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_collection_settings", return_value=default_settings
            ) as mock_get_settings,
        ):
            # --- Act --- #
            input_data = CreateCollectionInput(collection_name=collection_name)
            result = await _create_collection_impl(input_data)

            # --- Assert --- #
            mock_validate_name.assert_called_once_with(collection_name)
            mock_get_client.assert_called_once()
            mock_get_ef.assert_called_once_with("default")  # Assumes default config
            mock_get_settings.assert_called_once()  # Ensure settings were fetched

            expected_metadata_passed = {f"chroma:setting:{k.replace(':', '_')}": v for k, v in default_settings.items()}
            mock_client_instance.create_collection.assert_called_once_with(
                name=collection_name,
                metadata=expected_metadata_passed,
                embedding_function=mock_ef_instance,
                get_or_create=False,
            )
            assert len(result) == 1
            assert result[0].type == "text"
            # Add assertions on result content if necessary

    @pytest.mark.asyncio
    # REMOVE fixtures, use local patching
    async def test_create_collection_falls_back_to_default_ef(self):
        """Test _create_collection_impl uses 'default' EF name (implicitly)."""
        collection_name = "test-fallback-ef"

        # --- Mocks --- #
        mock_collection = MagicMock()
        mock_collection.name = collection_name
        mock_collection.id = "uuid-123"
        default_settings = get_collection_settings()  # Get defaults
        stored_metadata = {f"chroma:setting:{k.replace(':', '_')}": v for k, v in default_settings.items()}
        mock_collection.metadata = stored_metadata
        mock_collection.count.return_value = 0

        mock_client_instance = MagicMock()
        mock_client_instance.create_collection.return_value = mock_collection

        mock_ef_instance = MockEmbeddingFunction(name="default")

        # --- Patching --- #
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate_name,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_chroma_client", return_value=mock_client_instance
            ) as mock_get_client,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_embedding_function", return_value=mock_ef_instance
            ) as mock_get_ef,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_collection_settings", return_value=default_settings
            ) as mock_get_settings,
        ):
            # --- Act --- #
            input_data = CreateCollectionInput(collection_name=collection_name)
            await _create_collection_impl(input_data)

            # --- Assert --- #
            mock_validate_name.assert_called_once_with(collection_name)
            mock_get_client.assert_called_once()
            mock_get_ef.assert_called_once_with("default")  # Main assertion: default was used
            mock_get_settings.assert_called_once()

            expected_metadata_passed = {f"chroma:setting:{k.replace(':', '_')}": v for k, v in default_settings.items()}
            mock_client_instance.create_collection.assert_called_once_with(
                name=collection_name,
                metadata=expected_metadata_passed,
                embedding_function=mock_ef_instance,
                get_or_create=False,
            )

    @pytest.mark.asyncio
    # REMOVE fixtures, use local patching
    async def test_create_collection_success(self):
        """Test successful collection creation."""
        collection_name = "test_create_new"
        mock_collection_id = str(uuid.uuid4())

        # --- Mocks --- #
        mock_collection = MagicMock()
        mock_collection.name = collection_name
        mock_collection.id = mock_collection_id
        default_settings = get_collection_settings()  # Get defaults
        metadata_stored_by_chroma = {f"chroma:setting:{k.replace(':', '_')}": v for k, v in default_settings.items()}
        mock_collection.metadata = metadata_stored_by_chroma
        mock_collection.count.return_value = 0

        mock_client_instance = MagicMock()
        mock_client_instance.create_collection.return_value = mock_collection

        mock_ef_instance = MockEmbeddingFunction(name="default")

        # --- Patching --- #
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate_name,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_chroma_client", return_value=mock_client_instance
            ) as mock_get_client,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_embedding_function", return_value=mock_ef_instance
            ) as mock_get_ef,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_collection_settings", return_value=default_settings
            ) as mock_get_settings,
        ):
            # --- Act --- #
            input_model = CreateCollectionInput(collection_name=collection_name)
            result_list = await _create_collection_impl(input_model)

            # --- Assert --- #
            mock_validate_name.assert_called_once_with(collection_name)
            mock_get_client.assert_called_once()
            mock_get_ef.assert_called_once_with("default")
            mock_get_settings.assert_called_once()

            mock_client_instance.create_collection.assert_called_once()
            call_args, call_kwargs = mock_client_instance.create_collection.call_args
            assert call_kwargs["name"] == collection_name
            assert call_kwargs["metadata"] == metadata_stored_by_chroma
            assert call_kwargs["get_or_create"] is False
            assert call_kwargs["embedding_function"] is mock_ef_instance

            result_data = assert_successful_json_result(result_list)
            assert result_data.get("name") == collection_name
            assert result_data.get("id") == mock_collection_id
            assert "metadata" in result_data
            assert "settings" in result_data["metadata"]
            expected_settings_with_colons = {k.replace("_", ":"): v for k, v in default_settings.items()}
            assert result_data["metadata"]["settings"] == expected_settings_with_colons
            assert result_data.get("count") == 0

    @pytest.mark.asyncio
    # REMOVE mock_chroma_client fixture
    # No fixtures needed beyond class level
    async def test_create_collection_invalid_name(self):
        """Test collection name validation failure within the implementation."""
        invalid_name = "invalid-"
        error_msg = "Invalid collection name"

        # Patch validator and client method (to check not called)
        with (
            patch(
                "src.chroma_mcp.tools.collection_tools.validate_collection_name", side_effect=ValidationError(error_msg)
            ) as mock_validate_name,
            patch("chromadb.api.client.Client.create_collection") as mock_create_collection,
        ):
            # --- Act & Assert ---
            input_model = CreateCollectionInput(collection_name=invalid_name)
            with assert_raises_mcp_error(f"Validation Error: {error_msg}"):
                await _create_collection_impl(input_model)

            mock_validate_name.assert_called_once_with(invalid_name)
            mock_create_collection.assert_not_called()

    # --- _peek_collection_impl Tests ---
    @pytest.mark.asyncio
    # REMOVE Fixtures - Use local patching
    async def test_peek_collection_success(self):
        """Test successful peeking into a collection with a specific limit."""
        collection_name = "test_peek_exists"
        limit = 3
        expected_peek_result = {
            "ids": ["id1", "id2", "id3"],
            "documents": ["doc1", "doc2", "doc3"],
            "metadatas": [{"m": 1}, {"m": 2}, {"m": 3}],
            "embeddings": None,
        }

        # Local Mocks
        local_mock_collection = MagicMock()
        local_mock_collection.peek.return_value = expected_peek_result
        local_mock_client = MagicMock()
        local_mock_client.get_collection.return_value = local_mock_collection

        # Patch locally
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_chroma_client", return_value=local_mock_client
            ) as mock_get_client,
        ):
            input_model = PeekCollectionInput(collection_name=collection_name, limit=limit)
            result = await _peek_collection_impl(input_model)

            # Assertions
            mock_validate.assert_called_once_with(collection_name)
            mock_get_client.assert_called_once()
            local_mock_client.get_collection.assert_called_once_with(name=collection_name)
            local_mock_collection.peek.assert_called_once_with(limit=limit)
            assert_successful_json_result(result, expected_peek_result)

    @pytest.mark.asyncio
    # REMOVE Fixtures - Use local patching
    async def test_peek_collection_success_default_limit(self):
        """Test successful peeking using the default limit (10)."""
        collection_name = "test_peek_default"
        expected_peek_result = {"ids": ["default_id"]}

        # Local Mocks
        local_mock_collection = MagicMock()
        local_mock_collection.peek.return_value = expected_peek_result
        local_mock_client = MagicMock()
        local_mock_client.get_collection.return_value = local_mock_collection

        # Patch locally
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_chroma_client", return_value=local_mock_client
            ) as mock_get_client,
        ):
            input_model = PeekCollectionInput(collection_name=collection_name)  # Use default limit
            result = await _peek_collection_impl(input_model)

            # Assertions
            mock_validate.assert_called_once_with(collection_name)
            mock_get_client.assert_called_once()
            local_mock_client.get_collection.assert_called_once_with(name=collection_name)
            local_mock_collection.peek.assert_called_once_with(limit=10)  # Check default limit value
            assert_successful_json_result(result, expected_peek_result)

    @pytest.mark.asyncio
    # REMOVE Fixtures - Use local patching
    async def test_peek_collection_validation_error(self):
        """Test collection name validation failure for peek."""
        invalid_name = "peek-invalid--"
        error_msg = "Bad name for peek"

        # Patch locally (only validator needed, client shouldn't be called)
        with (
            patch(
                "src.chroma_mcp.tools.collection_tools.validate_collection_name", side_effect=ValidationError(error_msg)
            ) as mock_validate,
            patch("src.chroma_mcp.tools.collection_tools.get_chroma_client") as mock_get_client,
        ):
            input_model = PeekCollectionInput(collection_name=invalid_name)
            # Expect ValidationError message format
            with assert_raises_mcp_error(f"Validation Error: {error_msg}"):
                await _peek_collection_impl(input_model)

            # Assertions
            mock_validate.assert_called_once_with(invalid_name)
            mock_get_client.assert_not_called()  # Client should not be fetched if validation fails

    @pytest.mark.asyncio
    # REMOVE Fixtures - Use local patching
    async def test_peek_collection_not_found_error(self):
        """Test peek when the collection is not found (error from get_collection)."""
        collection_name = "peek_not_found"
        error_msg = f"Collection {collection_name} does not exist."

        # Local mock client configured to raise error
        local_mock_client = MagicMock()
        local_mock_client.get_collection.side_effect = ValueError(error_msg)

        # Patch locally
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_chroma_client", return_value=local_mock_client
            ) as mock_get_client,
        ):
            input_model = PeekCollectionInput(collection_name=collection_name)
            # Expect the specific 'not found' message from the ValueError handler
            with assert_raises_mcp_error(f"Tool Error: Collection '{collection_name}' not found."):
                await _peek_collection_impl(input_model)

            # Assertions
            mock_validate.assert_called_once_with(collection_name)
            mock_get_client.assert_called_once()
            local_mock_client.get_collection.assert_called_once_with(name=collection_name)

    @pytest.mark.asyncio
    # REMOVE Fixtures - Use local patching
    async def test_peek_collection_get_other_value_error(self):
        """Test other ValueError from get_collection during peek."""
        collection_name = "peek_get_val_err"
        error_msg = "Get value error during peek"

        # Local mock client configured to raise error
        local_mock_client = MagicMock()
        local_mock_client.get_collection.side_effect = ValueError(error_msg)

        # Patch locally
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_chroma_client", return_value=local_mock_client
            ) as mock_get_client,
        ):
            input_model = PeekCollectionInput(collection_name=collection_name)
            # Expect the specific 'Problem accessing collection' message from the ValueError handler
            with assert_raises_mcp_error(
                f"Tool Error: Problem accessing collection '{collection_name}'. Details: {error_msg}"
            ):
                await _peek_collection_impl(input_model)

            # Assertions
            mock_validate.assert_called_once_with(collection_name)
            mock_get_client.assert_called_once()
            local_mock_client.get_collection.assert_called_once_with(name=collection_name)

    @pytest.mark.asyncio
    # REMOVE Fixtures - Use local patching
    async def test_peek_collection_peek_exception(self):
        """Test handling of exception during the actual collection.peek call."""
        collection_name = "peek_itself_fails"
        error_msg = "Peek failed internally"

        # Local Mocks
        local_mock_collection = MagicMock()
        local_mock_collection.peek.side_effect = Exception(error_msg)
        local_mock_client = MagicMock()
        local_mock_client.get_collection.return_value = local_mock_collection

        # Patch Locally
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_chroma_client", return_value=local_mock_client
            ) as mock_get_client,
        ):
            input_model = PeekCollectionInput(collection_name=collection_name)
            # Expect the generic exception handler's message format
            expected_error_msg = (
                f"Tool Error: An unexpected error occurred peeking collection '{collection_name}'. Details: {error_msg}"
            )
            with assert_raises_mcp_error(expected_error_msg):
                await _peek_collection_impl(input_model)

            # Assertions
            mock_validate.assert_called_once_with(collection_name)
            mock_get_client.assert_called_once()
            local_mock_client.get_collection.assert_called_once_with(name=collection_name)
            # Check default limit for PeekCollectionInput is 10
            local_mock_collection.peek.assert_called_once_with(limit=10)

    # --- _create_collection_with_metadata_impl Tests ---
    @pytest.mark.asyncio
    # Remove all fixtures except reset_chroma_client_cache (autouse=True)
    @pytest.mark.usefixtures()
    async def test_create_collection_with_metadata_success(self):  # No fixture arguments needed
        """Test successful creation using the _with_metadata variant (as JSON string)."""
        collection_name = "test_create_with_meta_success"
        custom_metadata_dict = {"user_key": "user_value"}
        custom_metadata_json = json.dumps(custom_metadata_dict)
        mock_collection_id = str(uuid.uuid4())

        # --- Setup Local Mocks ---
        # Mock for the collection object that create_collection will return
        created_collection_mock = MagicMock()
        created_collection_mock.name = collection_name
        created_collection_mock.id = mock_collection_id
        created_collection_mock.metadata = custom_metadata_dict
        created_collection_mock.count.return_value = 0

        # Local mock for the Chroma client instance
        local_mock_client = MagicMock()
        local_mock_client.create_collection = MagicMock(return_value=created_collection_mock)

        # Define dummy EF class and instance
        class MockEF:
            def __call__(self, input):
                return [[0.0] * 10 for _ in input]

        local_mock_ef_instance = MockEF()

        # Local mock for the logger instance
        local_mock_logger_instance = MagicMock()

        input_model = CreateCollectionWithMetadataInput(collection_name=collection_name, metadata=custom_metadata_json)

        # --- Patch all dependencies locally ---
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as local_mock_validate,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_embedding_function", return_value=local_mock_ef_instance
            ) as local_mock_get_ef,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_logger", return_value=local_mock_logger_instance
            ) as mock_get_logger_call,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_chroma_client", return_value=local_mock_client
            ) as local_mock_get_client,
        ):
            # Note: We still rely on get_server_config implicitly inside the code for the default EF name

            # --- Act --- (Call the implementation)
            result_list = await _create_collection_with_metadata_impl(input_model)

            # --- Assert Internal Calls ---
            local_mock_validate.assert_called_once_with(collection_name)
            local_mock_get_ef.assert_called_once_with("default")  # Assuming default EF name
            mock_get_logger_call.assert_called_once()
            local_mock_get_client.assert_called_once()  # Ensure client getter was called

            # Assert Logger Call (should pass)
            expected_log_message = f"Attempting to create collection '{collection_name}' with provided parsed metadata: {custom_metadata_dict}"
            local_mock_logger_instance.info.assert_any_call(expected_log_message)
            assert local_mock_logger_instance.error.call_count == 0

        # --- Assert Actual Client Call on the LOCAL mock ---
        local_mock_client.create_collection.assert_called_once()
        # Check args passed to the local client mock
        call_args = local_mock_client.create_collection.call_args
        assert call_args.kwargs["name"] == collection_name
        assert call_args.kwargs["metadata"] == custom_metadata_dict
        assert call_args.kwargs["embedding_function"] is local_mock_ef_instance

        # --- Assert Result ---
        result_data = assert_successful_json_result(result_list)
        assert result_data.get("name") == collection_name
        assert result_data.get("id") == mock_collection_id

    @pytest.mark.asyncio
    # Patch create_collection via decorator, use local patch for validator
    @patch("chromadb.api.client.Client.create_collection")
    @pytest.mark.usefixtures("mock_chroma_client", "mock_server_config", "mock_embedding_functions")
    async def test_create_collection_with_metadata_validation_error(
        self, mock_create_collection, mock_chroma_client, mock_server_config, mock_embedding_functions
    ):
        """Test create with metadata validation error using local patch for validator."""
        invalid_name = "meta-invalid--"
        error_msg = "Bad name from local patch"
        valid_metadata_json = '{"key": "value"}'

        # Patch the validator locally where it's imported in the implementation module
        with patch(
            "src.chroma_mcp.tools.collection_tools.validate_collection_name", side_effect=ValidationError(error_msg)
        ) as local_mock_validate:
            input_model = CreateCollectionWithMetadataInput(collection_name=invalid_name, metadata=valid_metadata_json)
            with assert_raises_mcp_error(f"Validation Error: {error_msg}"):
                await _create_collection_with_metadata_impl(input_model)

            local_mock_validate.assert_called_once_with(invalid_name)
            mock_create_collection.assert_not_called()  # Client method should not be called

    @pytest.mark.asyncio
    async def test_create_collection_with_metadata_invalid_json(self):
        """Test create with metadata when the metadata string is invalid JSON."""
        collection_name = "valid_json_test_coll"
        invalid_json_string = '{"key": "value", "unterminated'
        expected_error_substring = "Invalid JSON format for metadata field"  # More general check

        # Patch only the client method to check it's not called
        with patch("chromadb.api.client.Client.create_collection") as mock_create_collection:
            input_model = CreateCollectionWithMetadataInput(
                collection_name=collection_name, metadata=invalid_json_string
            )
            with assert_raises_mcp_error(expected_error_substring):
                await _create_collection_with_metadata_impl(input_model)
            mock_create_collection.assert_not_called()

    @pytest.mark.asyncio
    async def test_create_collection_with_metadata_already_exists(self):
        """Test create with metadata when collection already exists (ValueError from client)."""
        collection_name = "meta_exists_test"
        error_msg = f"Collection {collection_name} already exists."
        valid_metadata_json = '{"key": "value"}'

        # Patch validator and get_chroma_client
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate_name,
            patch("src.chroma_mcp.tools.collection_tools.get_chroma_client") as mock_get_client,
        ):
            # Configure the mock client returned by get_chroma_client
            mock_client_instance = MagicMock()
            # Set the side effect directly on the mock client's method
            mock_client_instance.create_collection.side_effect = ValueError(error_msg)
            mock_get_client.return_value = mock_client_instance

            input_model = CreateCollectionWithMetadataInput(
                collection_name=collection_name, metadata=valid_metadata_json
            )
            # Check for the specific error message wrapped by the implementation
            with assert_raises_mcp_error(f"Tool Error: Collection '{collection_name}' already exists."):
                await _create_collection_with_metadata_impl(input_model)

            mock_validate_name.assert_called_once_with(collection_name)
            # Assert the create_collection method was called on the mock client instance
            mock_client_instance.create_collection.assert_called_once()

    @pytest.mark.asyncio
    async def test_create_collection_with_metadata_unexpected_error(self):
        """Test create with metadata handling unexpected errors from client."""
        collection_name = "meta_unexpected_err_test"
        error_msg = "Something else failed badly"
        valid_metadata_json = '{"key": "value"}'

        # Patch validator and get_chroma_client
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate_name,
            patch("src.chroma_mcp.tools.collection_tools.get_chroma_client") as mock_get_client,
        ):
            # Configure the mock client returned by get_chroma_client
            mock_client_instance = MagicMock()
            # Set the side effect directly on the mock client's method
            mock_client_instance.create_collection.side_effect = Exception(error_msg)
            mock_get_client.return_value = mock_client_instance

            input_model = CreateCollectionWithMetadataInput(
                collection_name=collection_name, metadata=valid_metadata_json
            )
            # Check for the generic error message wrapped by the implementation
            with assert_raises_mcp_error(
                f"Tool Error: An unexpected error occurred while creating collection '{collection_name}'. Details: {error_msg}"
            ):
                await _create_collection_with_metadata_impl(input_model)

            mock_validate_name.assert_called_once_with(collection_name)
            # Assert the create_collection method was called on the mock client instance
            mock_client_instance.create_collection.assert_called_once()

    # --- _list_collections_impl Tests ---
    @pytest.mark.asyncio
    # REMOVE Fixture - Use local patching
    async def test_list_collections_success_defaults(self):
        """Test successful default collection listing (no filters, no pagination)."""
        all_names = ["coll_a", "coll_b"]

        # Local Mock
        local_mock_client = MagicMock()
        local_mock_client.list_collections.return_value = all_names

        # Patch locally
        with patch(
            "src.chroma_mcp.tools.collection_tools.get_chroma_client", return_value=local_mock_client
        ) as mock_get_client:
            input_model = ListCollectionsInput()
            result_list = await _list_collections_impl(input_model)

            # Assertions
            mock_get_client.assert_called_once()
            local_mock_client.list_collections.assert_called_once()
            local_mock_client.get_collection.assert_not_called()  # Still shouldn't be called

            result_data = assert_successful_json_result(result_list)
            assert result_data.get("collection_names") == all_names
            assert result_data.get("total_count") == len(all_names)
            assert result_data.get("limit") == 0
            assert result_data.get("offset") == 0

    @pytest.mark.asyncio
    # REMOVE Fixture - Use local patching
    async def test_list_collections_with_filter_pagination(self):
        """Test listing with name filter and pagination."""
        all_names = ["apple", "banana", "apricot", "avocado"]

        # Local Mock
        local_mock_client = MagicMock()
        local_mock_client.list_collections.return_value = all_names

        # Patch locally
        with patch(
            "src.chroma_mcp.tools.collection_tools.get_chroma_client", return_value=local_mock_client
        ) as mock_get_client:
            input_model = ListCollectionsInput(limit=1, offset=1, name_contains="ap")
            result_list = await _list_collections_impl(input_model)

            # Assertions
            mock_get_client.assert_called_once()
            local_mock_client.list_collections.assert_called_once()
            local_mock_client.get_collection.assert_not_called()

            result_data = assert_successful_json_result(result_list)
            assert result_data.get("collection_names") == ["apricot"]
            assert result_data.get("total_count") == 2
            assert result_data.get("limit") == 1
            assert result_data.get("offset") == 1

    @pytest.mark.skip(
        reason="got empty parameter set ['limit', 'offset', 'expected_error_msg'], function test_list_collections_validation_error at /Users/dominikus/git/nold-ai/chroma_mcp_server/tests/tools/test_collection_tools.py:745"
    )
    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "limit, offset, expected_error_msg",
        [
            # (-1, 0, "limit cannot be negative"), # Pydantic handles this
            # (0, -1, "offset cannot be negative"), # Pydantic handles this
        ],
        ids=[
            # "negative_limit", # Removed as Pydantic validation
            # "negative_offset", # Removed as Pydantic validation
        ],
    )
    # Remove fixture use here, test only Pydantic validation
    async def test_list_collections_validation_error(
        self, limit, offset, expected_error_msg  # Remove mock_chroma_client_collections
    ):
        """Test input validation for limit and offset in list_collections."""
        # This test correctly uses Pydantic's validation, but we need one for client errors
        # No specific validation error raised IN the function, Pydantic handles it.
        # Instead, we add a test for client-side errors.
        # REMOVED the original test body as Pydantic handles this.
        # The assert_raises_mcp_error needs to be outside the await for Pydantic validation
        with pytest.raises(Exception):  # Pydantic raises its own ValidationError
            _ = ListCollectionsInput(limit=limit, offset=offset)
        # Keep the assertion structure for documentation, though it might not be hit
        # with assert_raises_mcp_error(expected_error_msg):
        #     input_model = ListCollectionsInput(limit=limit, offset=offset)
        #     _ = await _list_collections_impl(input_model)

    @pytest.mark.asyncio
    # Patch the actual client method to raise error
    async def test_list_collections_client_error(self):
        """Test handling of errors during client.list_collections."""
        error_msg = "Client connection failed"

        # Patch list_collections on the *mock client* returned by get_chroma_client
        with patch("src.chroma_mcp.tools.collection_tools.get_chroma_client") as mock_get_client:
            # Configure the mock client
            mock_client_instance = MagicMock()
            mock_client_instance.list_collections.side_effect = Exception(error_msg)
            mock_get_client.return_value = mock_client_instance

            input_model = ListCollectionsInput()
            with assert_raises_mcp_error(f"Tool Error: Error listing collections. Details: {error_msg}"):
                await _list_collections_impl(input_model)
            # Assert the method was called on the mock client
            mock_client_instance.list_collections.assert_called_once()

    # --- _get_collection_impl Tests ---
    @pytest.mark.asyncio
    # REMOVE Fixtures - Use local patching
    async def test_get_collection_success(self):
        """Test getting existing collection info."""
        collection_name = "my_coll"
        mock_collection_id = "test-id-123"
        mock_metadata_stored = {"description": "test desc", "chroma:setting:hnsw_space": "l2"}
        mock_count = 42
        # Ensure embeddings are None in the peek result if not expected
        mock_peek = {"ids": ["p1"], "documents": ["peek doc"], "embeddings": None}

        # Local Mocks
        local_mock_collection = MagicMock()
        local_mock_collection.name = collection_name
        local_mock_collection.id = mock_collection_id
        local_mock_collection.metadata = mock_metadata_stored
        local_mock_collection.count.return_value = mock_count
        local_mock_collection.peek.return_value = mock_peek

        local_mock_client = MagicMock()
        local_mock_client.get_collection.return_value = local_mock_collection

        # Patch locally
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_chroma_client", return_value=local_mock_client
            ) as mock_get_client,
        ):
            input_model = GetCollectionInput(collection_name=collection_name)
            result = await _get_collection_impl(input_model)

            # Assertions
            mock_validate.assert_called_once_with(collection_name)
            mock_get_client.assert_called_once()
            local_mock_client.get_collection.assert_called_once_with(name=collection_name)
            local_mock_collection.count.assert_called_once()
            local_mock_collection.peek.assert_called_once_with(limit=5)

            result_data = assert_successful_json_result(result)
            assert result_data.get("name") == collection_name
            assert result_data.get("id") == mock_collection_id
            assert result_data.get("count") == mock_count
            assert result_data.get("metadata") == _reconstruct_metadata(mock_metadata_stored)
            assert result_data.get("sample_entries") == mock_peek

    @pytest.mark.asyncio
    # REMOVE Fixtures - Use local patching
    async def test_get_collection_not_found(self):
        """Test getting a non-existent collection (handled in impl)."""
        collection_name = "not_found_coll"
        error_message = f"Collection {collection_name} does not exist."

        # Local Mock Client
        local_mock_client = MagicMock()
        # Configure get_collection to raise the specific not found error
        local_mock_client.get_collection.side_effect = ValueError(error_message)

        # Patch locally
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_chroma_client", return_value=local_mock_client
            ) as mock_get_client,
        ):
            input_model = GetCollectionInput(collection_name=collection_name)
            # Expect the specific error message from the ValueError handler
            with assert_raises_mcp_error(f"Tool Error: Collection '{collection_name}' not found."):
                await _get_collection_impl(input_model)

            # Assertions
            mock_validate.assert_called_once_with(collection_name)
            mock_get_client.assert_called_once()
            local_mock_client.get_collection.assert_called_once_with(name=collection_name)

    @pytest.mark.asyncio
    # REMOVE Fixtures - Use local patching
    async def test_get_collection_unexpected_error(self):
        """Test handling of unexpected error during get collection."""
        collection_name = "test_get_fail"
        error_msg = "Get failed unexpectedly"

        # Local Mock Client
        local_mock_client = MagicMock()
        # Configure get_collection to raise a generic Exception
        local_mock_client.get_collection.side_effect = Exception(error_msg)

        # Patch locally
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_chroma_client", return_value=local_mock_client
            ) as mock_get_client,
        ):
            input_model = GetCollectionInput(collection_name=collection_name)
            # Expect the generic error message from the final except block
            expected_error = f"Tool Error: An unexpected error occurred while getting collection '{collection_name}'. Details: {error_msg}"
            with assert_raises_mcp_error(expected_error):
                await _get_collection_impl(input_model)

            # Assertions
            mock_validate.assert_called_once_with(collection_name)
            mock_get_client.assert_called_once()
            local_mock_client.get_collection.assert_called_once_with(name=collection_name)

    @pytest.mark.asyncio
    # REMOVE Fixtures - Use local patching
    async def test_get_collection_validation_error(self):
        """Test handling of validation error for collection name."""
        collection_name = "invalid--name"
        error_msg = "Invalid name format"

        # Patch locally (validator raises, client getter shouldn't be called)
        with (
            patch(
                "src.chroma_mcp.tools.collection_tools.validate_collection_name", side_effect=ValidationError(error_msg)
            ) as mock_validate,
            patch("src.chroma_mcp.tools.collection_tools.get_chroma_client") as mock_get_client,
        ):
            input_model = GetCollectionInput(collection_name=collection_name)
            # Expect the specific Validation Error message
            with assert_raises_mcp_error(f"Validation Error: {error_msg}"):
                await _get_collection_impl(input_model)

            # Assertions
            mock_validate.assert_called_once_with(collection_name)
            mock_get_client.assert_not_called()  # get_chroma_client should not be called

    @pytest.mark.asyncio
    # REMOVE Fixtures - Use local patching
    async def test_get_collection_other_value_error(self):
        """Test handling of other ValueErrors from get_collection."""
        collection_name = "test_get_value_err"
        error_msg = "Another value error during get"

        # Local Mock Client
        local_mock_client = MagicMock()
        # Configure get_collection to raise a different ValueError
        local_mock_client.get_collection.side_effect = ValueError(error_msg)

        # Patch locally
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_chroma_client", return_value=local_mock_client
            ) as mock_get_client,
        ):
            input_model = GetCollectionInput(collection_name=collection_name)
            # Expect the specific 'Invalid parameter' message from the ValueError handler
            expected_error = f"Tool Error: Invalid parameter getting collection. Details: {error_msg}"
            with assert_raises_mcp_error(expected_error):
                await _get_collection_impl(input_model)

            # Assertions
            mock_validate.assert_called_once_with(collection_name)
            mock_get_client.assert_called_once()
            local_mock_client.get_collection.assert_called_once_with(name=collection_name)

    @pytest.mark.asyncio
    # REMOVE Fixtures - Use local patching
    async def test_get_collection_peek_error(self):
        """Test that get_collection returns info even if peek fails."""
        collection_name = "test_peek_fail"
        mock_collection_id = "peek-fail-id"
        mock_metadata_stored = {"desc": "test"}
        mock_count = 5
        peek_error_msg = "Peek failed"

        # Local Mocks
        local_mock_collection = MagicMock()
        local_mock_collection.name = collection_name
        local_mock_collection.id = mock_collection_id
        local_mock_collection.metadata = mock_metadata_stored
        local_mock_collection.count.return_value = mock_count
        # Configure peek to raise an error
        local_mock_collection.peek.side_effect = Exception(peek_error_msg)

        local_mock_client = MagicMock()
        # Configure get_collection to succeed and return the collection mock
        local_mock_client.get_collection.return_value = local_mock_collection

        # Patch locally
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_chroma_client", return_value=local_mock_client
            ) as mock_get_client,
        ):
            input_model = GetCollectionInput(collection_name=collection_name)
            result_list = await _get_collection_impl(input_model)

            # Assertions on internal calls
            mock_validate.assert_called_once_with(collection_name)
            mock_get_client.assert_called_once()
            local_mock_client.get_collection.assert_called_once_with(name=collection_name)
            local_mock_collection.count.assert_called_once()
            local_mock_collection.peek.assert_called_once_with(limit=5)

            # Assert success, but check the sample_entries field for the error
            result_data = assert_successful_json_result(result_list)
            assert result_data["name"] == collection_name
            assert result_data["count"] == mock_count
            assert "sample_entries" in result_data
            assert isinstance(result_data["sample_entries"], dict)
            assert "error" in result_data["sample_entries"]
            assert peek_error_msg in result_data["sample_entries"]["error"]
            assert result_data.get("id") == mock_collection_id  # Add check for ID
            assert result_data.get("metadata") == _reconstruct_metadata(mock_metadata_stored)  # Add check for metadata

    # --- _rename_collection_impl Tests ---
    @pytest.mark.asyncio
    # Remove fixture, patch manually inside
    # @pytest.mark.usefixtures("mock_validate_name")
    async def test_rename_collection_success(self):
        """Test successful collection renaming."""
        original_name = "rename_me"
        new_name = "renamed_successfully"

        # Patch validator and get_chroma_client
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate_name,
            patch("src.chroma_mcp.tools.collection_tools.get_chroma_client") as mock_get_client,
        ):
            # Configure mock client and the collection it returns
            mock_client_instance = MagicMock()
            mock_collection_instance = MagicMock()
            mock_collection_instance.modify.return_value = None  # modify returns None on success
            mock_client_instance.get_collection.return_value = mock_collection_instance
            mock_get_client.return_value = mock_client_instance

            # --- Act ---
            input_model = RenameCollectionInput(collection_name=original_name, new_name=new_name)
            result = await _rename_collection_impl(input_model)

            # --- Assert ---
            mock_validate_name.assert_has_calls([call(original_name), call(new_name)])
            mock_get_client.assert_called_once()
            mock_client_instance.get_collection.assert_called_once_with(name=original_name)
            mock_collection_instance.modify.assert_called_once_with(name=new_name)

            # Assert successful result message
            assert isinstance(result, list)
            assert len(result) == 1
            assert isinstance(result[0], types.TextContent)
            assert f"Collection '{original_name}' successfully renamed to '{new_name}'." in result[0].text

    @pytest.mark.asyncio
    # Remove fixture, patch manually inside
    # @pytest.mark.usefixtures("mock_validate_name")
    async def test_rename_collection_original_not_found(self):
        """Test renaming when the original collection does not exist."""
        original_name = "original_not_found"
        new_name = "new_name_irrelevant"

        # Patch validator and get_chroma_client
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate_name,
            patch("src.chroma_mcp.tools.collection_tools.get_chroma_client") as mock_get_client,
        ):
            # Configure the mock client's get_collection method to raise the error
            mock_client_instance = MagicMock()
            mock_client_instance.get_collection.side_effect = ValueError(f"Collection {original_name} does not exist.")
            mock_get_client.return_value = mock_client_instance

            # --- Act ---
            input_model = RenameCollectionInput(collection_name=original_name, new_name=new_name)
            with assert_raises_mcp_error(f"Tool Error: Collection '{original_name}' not found."):
                await _rename_collection_impl(input_model)

            # --- Assert ---
            mock_validate_name.assert_has_calls([call(original_name), call(new_name)])
            # Assert the get_collection method was called on the mock client
            mock_client_instance.get_collection.assert_called_once_with(name=original_name)

    @pytest.mark.asyncio
    # Remove fixture, patch manually inside
    # @pytest.mark.usefixtures("mock_validate_name")
    async def test_rename_collection_new_name_exists(self):
        """Test renaming when the new name already exists."""
        original_name = "original_exists"
        new_name = "new_name_exists"

        # Patch validator, client.get_collection, and collection.modify
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate_name,
            patch("chromadb.api.client.Client.get_collection") as mock_get_collection_method,
            patch(
                "chromadb.api.models.Collection.Collection.modify",  # Patch the real modify
                side_effect=ValueError(f"Collection {new_name} already exists."),
            ) as mock_modify_method,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_chroma_client", return_value=MagicMock()
            ) as mock_get_client,
        ):  # Add this patch
            # Configure the mock collection returned by get_collection
            mock_collection_instance = MagicMock()

    @pytest.mark.asyncio
    # Remove fixture, patch manually inside
    # @pytest.mark.usefixtures("mock_validate_name")
    async def test_rename_collection_unexpected_error(self):
        """Test unexpected error during rename (e.g., during modify)."""
        original_name = "rename_fail_orig"
        new_name = "rename_fail_new"
        error_msg = "Rename blew up"

        # Patch validator, client.get_collection, and collection.modify
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate_name,
            patch("chromadb.api.client.Client.get_collection") as mock_get_collection_method,
            patch(
                "chromadb.api.models.Collection.Collection.modify",  # Patch the real modify
                side_effect=Exception(error_msg),
            ) as mock_modify_method,
            patch(
                "src.chroma_mcp.tools.collection_tools.get_chroma_client", return_value=MagicMock()
            ) as mock_get_client,
        ):  # Add this patch
            # Configure the mock collection returned by get_collection
            mock_collection_instance = MagicMock()

    @pytest.mark.asyncio
    # Remove fixture, patch manually inside
    # @pytest.mark.usefixtures("mock_validate_name")
    async def test_rename_collection_other_value_error(self):
        """Test handling of other ValueErrors during rename (e.g., from modify)."""
        original_name = "rename_val_err_orig"
        new_name = "rename_val_err_new"
        error_msg = "Some other value issue"

        # Patch validator AND get_chroma_client
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate_name,
            patch("src.chroma_mcp.tools.collection_tools.get_chroma_client") as mock_get_chroma_client,
        ):
            # Configure mock client and the collection it returns
            mock_client_instance = MagicMock()
            mock_collection_instance = MagicMock()
            mock_collection_instance.modify.side_effect = ValueError(error_msg)
            mock_client_instance.get_collection.return_value = mock_collection_instance

            # Make get_chroma_client return our mock client
            mock_get_chroma_client.return_value = mock_client_instance

            input_model = RenameCollectionInput(collection_name=original_name, new_name=new_name)
            # Expect the ValueError block (the one that doesn't check specific messages)
            with assert_raises_mcp_error(f"Invalid parameter during rename. Details: {error_msg}"):
                await _rename_collection_impl(input_model)

            # Assertions
            mock_validate_name.assert_any_call(original_name)
            mock_validate_name.assert_any_call(new_name)
            mock_get_chroma_client.assert_called_once()
            # Assert get_collection was called on the *mock* client
            mock_client_instance.get_collection.assert_called_once_with(name=original_name)
            # Assert modify was called on the *mock* collection
            mock_collection_instance.modify.assert_called_once_with(name=new_name)

    # --- _delete_collection_impl Tests ---
    @pytest.mark.asyncio
    # REMOVE explicit fixture if applied at class level
    # @pytest.mark.usefixtures("mock_server_config")
    async def test_delete_collection_success(self):  # Keep self only
        """Test successful collection deletion."""
        collection_name = "delete_me"

        # Use nested patches for validator and get_chroma_client
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate,
            patch("src.chroma_mcp.tools.collection_tools.get_chroma_client") as mock_get_chroma_client,
        ):
            # Configure mock client and its delete method
            mock_client_instance = MagicMock()
            mock_client_instance.delete_collection.return_value = None  # Successful delete
            mock_get_chroma_client.return_value = mock_client_instance

            # --- Act ---
            input_model = DeleteCollectionInput(collection_name=collection_name)
            result = await _delete_collection_impl(input_model)

            # --- Assert ---
            mock_validate.assert_called_once_with(collection_name)
            mock_get_chroma_client.assert_called_once()
            # Assert delete was called on the mock client instance
            mock_client_instance.delete_collection.assert_called_once_with(name=collection_name)
            # Check plain text result
            assert isinstance(result, list)
            assert len(result) == 1
            assert isinstance(result[0], types.TextContent)
            assert result[0].type == "text"
            assert result[0].text == f"Collection '{collection_name}' deleted successfully."
            # assert_successful_json_result(result, {"message": f"Collection '{collection_name}' deleted successfully."})

    @pytest.mark.asyncio
    # REMOVE explicit fixture if applied at class level
    # @pytest.mark.usefixtures("mock_server_config")
    async def test_delete_collection_not_found(self):  # Keep self only
        """Test deleting a non-existent collection (error from delete_collection)."""
        collection_name = "not_found_delete"
        error_message = f"Collection {collection_name} does not exist."

        # Use nested patches for validator and get_chroma_client
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate,
            patch("src.chroma_mcp.tools.collection_tools.get_chroma_client") as mock_get_chroma_client,
        ):
            # Configure mock client to raise ValueError on delete
            mock_client_instance = MagicMock()
            mock_client_instance.delete_collection.side_effect = ValueError(error_message)
            mock_get_chroma_client.return_value = mock_client_instance

            # --- Act & Assert ---
            input_model = DeleteCollectionInput(collection_name=collection_name)
            with assert_raises_mcp_error(f"Tool Error: Collection '{collection_name}' not found."):
                await _delete_collection_impl(input_model)

            # Assert mocks
            mock_validate.assert_called_once_with(collection_name)
            mock_get_chroma_client.assert_called_once()
            mock_client_instance.delete_collection.assert_called_once_with(name=collection_name)

    @pytest.mark.asyncio
    # REMOVE explicit fixture if applied at class level
    # @pytest.mark.usefixtures("mock_server_config")
    async def test_delete_collection_unexpected_error(self):  # Keep self only
        """Test unexpected error during the delete_collection call."""
        collection_name = "delete_fail"
        error_msg = "Delete failed unexpectedly"

        # Use nested patches for validator and get_chroma_client
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate,
            patch("src.chroma_mcp.tools.collection_tools.get_chroma_client") as mock_get_chroma_client,
        ):
            # Configure mock client to raise generic Exception on delete
            mock_client_instance = MagicMock()
            mock_client_instance.delete_collection.side_effect = Exception(error_msg)
            mock_get_chroma_client.return_value = mock_client_instance

            input_model = DeleteCollectionInput(collection_name=collection_name)
            # Expect the generic Exception handler message
            with assert_raises_mcp_error(
                f"An unexpected error occurred deleting collection '{collection_name}'. Details: {error_msg}"
            ):
                await _delete_collection_impl(input_model)

            # Assert mocks
            mock_validate.assert_called_once_with(collection_name)
            mock_get_chroma_client.assert_called_once()
            mock_client_instance.delete_collection.assert_called_once_with(name=collection_name)

    @pytest.mark.asyncio
    # REMOVE explicit fixture if applied at class level
    # @pytest.mark.usefixtures("mock_server_config")
    async def test_delete_collection_validation_error(self):
        """Test validation error (e.g., invalid name) before client call."""
        collection_name = "invalid name!"
        error_msg = "Invalid collection name format"

        # Patch only the validator to raise error
        with (
            patch(
                "src.chroma_mcp.tools.collection_tools.validate_collection_name", side_effect=ValidationError(error_msg)
            ) as mock_validate,
            patch("src.chroma_mcp.tools.collection_tools.get_chroma_client") as mock_get_chroma_client,
        ):
            input_model = DeleteCollectionInput(collection_name=collection_name)
            with assert_raises_mcp_error(f"Validation Error: {error_msg}"):
                await _delete_collection_impl(input_model)

            # Assert mocks
            mock_validate.assert_called_once_with(collection_name)
            mock_get_chroma_client.assert_not_called()  # Client should not be fetched

    @pytest.mark.asyncio
    # REMOVE explicit fixture if applied at class level
    # @pytest.mark.usefixtures("mock_server_config")
    async def test_delete_collection_other_value_error(self):  # Keep self only
        """Test handling of other ValueErrors during the delete_collection call."""
        collection_name = "delete_val_err"
        error_msg = "Another delete value error"

        # Use nested patches for validator and get_chroma_client
        with (
            patch("src.chroma_mcp.tools.collection_tools.validate_collection_name") as mock_validate,
            patch("src.chroma_mcp.tools.collection_tools.get_chroma_client") as mock_get_chroma_client,
        ):
            # Configure mock client to raise specific ValueError (not 'not found')
            mock_client_instance = MagicMock()
            mock_client_instance.delete_collection.side_effect = ValueError(error_msg)
            mock_get_chroma_client.return_value = mock_client_instance

            input_model = DeleteCollectionInput(collection_name=collection_name)
            # Expect the specific ValueError handler message (check the impl for exact wording)
            with assert_raises_mcp_error(
                f"Invalid parameter during delete operation for '{collection_name}'. Details: {error_msg}"
            ):
                await _delete_collection_impl(input_model)

            # Assert mocks
            mock_validate.assert_called_once_with(collection_name)
            mock_get_chroma_client.assert_called_once()
            mock_client_instance.delete_collection.assert_called_once_with(name=collection_name)



"""
Sequential thinking tools for managing thought chains and context in ChromaDB.
"""

import time
import uuid
import json
import logging

from typing import Dict, List, Optional, Any
from dataclasses import dataclass, asdict

from mcp import types
from mcp.shared.exceptions import McpError
from mcp.types import ErrorData, INVALID_PARAMS, INTERNAL_ERROR
from pydantic import BaseModel, Field, field_validator, ConfigDict

# Import InvalidDimensionException
from chromadb.errors import InvalidDimensionException

from ..types import ThoughtMetadata
from ..utils import (
    get_logger,
    get_chroma_client,
    get_embedding_function,
    ValidationError,
    get_server_config,
)

# Constants
THOUGHTS_COLLECTION = "sequential_thoughts_v1"
SESSIONS_COLLECTION = "thinking_sessions"
DEFAULT_SIMILARITY_THRESHOLD = 0.6


@dataclass
class ThoughtMetadata:
    """Metadata structure for thoughts."""

    session_id: str
    thought_number: int
    total_thoughts: int
    timestamp: int
    branch_from_thought: Optional[int] = None
    branch_id: Optional[str] = None
    next_thought_needed: bool = False
    custom_data: Optional[Dict[str, Any]] = None


# Get logger instance for this module
# logger = get_logger("tools.thinking")

# --- Pydantic Input Models for Thinking Tools ---


class SequentialThinkingInput(BaseModel):
    thought: str = Field(..., description="Content of the thought being recorded.")
    thought_number: int = Field(..., gt=0, description="Sequential number of this thought (must be > 0).")
    total_thoughts: int = Field(..., description="Total anticipated number of thoughts in this sequence.")
    session_id: str = Field(default="", description="Unique session ID. If empty, a new session ID is generated.")
    branch_id: str = Field(
        default="", description="Optional identifier for a branch within the session. Empty if none."
    )
    branch_from_thought: int = Field(
        default=0,
        description="If creating a new branch, the parent thought number (> 0) it originates from. 0 if not branching.",
    )
    next_thought_needed: bool = Field(False, description="Flag indicating if a subsequent thought is expected.")

    model_config = ConfigDict(extra="forbid")


class SequentialThinkingWithCustomDataInput(SequentialThinkingInput):
    custom_data: str = Field(
        ..., description='Dictionary for arbitrary metadata as a JSON string (e.g., \'{"key": "value"}\').'
    )

    model_config = ConfigDict(extra="forbid")


class FindSimilarThoughtsInput(BaseModel):
    query: str = Field(..., description="Text to search for similar thoughts.")
    session_id: str = Field(
        default="", description="If provided, limits search to this session. Empty for global search."
    )
    n_results: int = Field(5, ge=1, description="Maximum number of similar thoughts to return (must be >= 1).")
    threshold: float = Field(
        default=-1.0,
        description="Similarity score threshold (0.0 to 1.0). Lower distance is more similar. -1.0 to use default.",
    )
    include_branches: bool = Field(True, description="Whether to include thoughts from branches in the search.")

    model_config = ConfigDict(extra="forbid")


class GetSessionSummaryInput(BaseModel):
    session_id: str = Field(..., description="The unique identifier for the thinking session to summarize.")
    include_branches: bool = Field(True, description="Whether to include thoughts from branches in the summary.")

    model_config = ConfigDict(extra="forbid")


class FindSimilarSessionsInput(BaseModel):
    query: str = Field(..., description="Text to search for similar thinking sessions based on overall content.")
    n_results: int = Field(5, ge=1, description="Maximum number of similar sessions to return (must be >= 1).")
    threshold: float = Field(
        default=-1.0,
        description="Similarity score threshold (0.0 to 1.0). Lower distance is more similar. -1.0 to use default.",
    )

    model_config = ConfigDict(extra="forbid")


# --- End Pydantic Input Models ---

# --- Implementation Functions ---


# Wrapper for the base variant (no custom_data)
async def _sequential_thinking_impl(input_data: SequentialThinkingInput) -> List[types.TextContent]:
    """Records a thought within a thinking session (without custom data).

    Args:
        input_data: A SequentialThinkingInput object containing validated arguments.

    Returns:
        A list containing a single TextContent object.
    """
    # Pass custom_data as None explicitly
    return await _base_sequential_thinking_impl(input_data, custom_data_json=None)


# Wrapper for the variant with custom_data
async def _sequential_thinking_with_custom_data_impl(
    input_data: SequentialThinkingWithCustomDataInput,
) -> List[types.TextContent]:
    """Records a thought within a thinking session (with custom data).

    Args:
        input_data: A SequentialThinkingWithCustomDataInput object containing validated arguments.

    Returns:
        A list containing a single TextContent object.
    """
    # Extract custom_data and pass it to the base implementation
    return await _base_sequential_thinking_impl(input_data, custom_data_json=input_data.custom_data)


# Base implementation (renamed from original _sequential_thinking_impl)
async def _base_sequential_thinking_impl(
    input_data: SequentialThinkingInput, custom_data_json: Optional[str]
) -> List[types.TextContent]:
    """Base implementation for recording a thought.

    Args:
        input_data: A SequentialThinkingInput (or subclass that conforms) object.
        custom_data_json: The custom data as a JSON string (can be None).

    Returns:
        A list containing a single TextContent object.
    """
    # Use root logger for entry debug
    logging.info("--- ENTERING _base_sequential_thinking_impl ---")

    logger = get_logger("tools.thinking")  # Keep module logger for other messages
    try:
        # Access validated data from input model
        thought = input_data.thought
        thought_number = input_data.thought_number
        total_thoughts = input_data.total_thoughts
        session_id = input_data.session_id  # Could be ""
        branch_id = input_data.branch_id  # Could be ""
        branch_from_thought = input_data.branch_from_thought  # Could be 0
        next_thought_needed = input_data.next_thought_needed  # Has default
        # custom_data_json is passed as an argument (remains Optional[str])

        # --- Parse Custom Data JSON --- #
        parsed_custom_data: Optional[Dict[str, Any]] = None
        if custom_data_json:
            try:
                parsed_custom_data = json.loads(custom_data_json)
                if not isinstance(parsed_custom_data, dict):
                    raise ValueError("Custom data string must decode to a JSON object (dictionary).")
            except json.JSONDecodeError as e:
                logger.warning(f"Failed to parse custom_data JSON string: {e}")
                raise McpError(
                    ErrorData(code=INVALID_PARAMS, message=f"Invalid JSON format for custom_data string: {str(e)}")
                )
            except ValueError as e:  # Catch the isinstance check
                logger.warning(f"Custom data did not decode to a dictionary: {e}")
                raise McpError(ErrorData(code=INVALID_PARAMS, message=str(e)))
        # --- End Parsing --- #

        # Use root logger before client calls
        logging.info("--- Getting Chroma client... ---")
        client = get_chroma_client()
        logging.info("--- Chroma client obtained. ---")

        # Get the server's default embedding function
        server_ef_name = get_server_config().embedding_function_name
        default_ef = get_embedding_function(server_ef_name)
        logger.debug(f"Using embedding function '{server_ef_name}' for thinking collections.")

        # Ensure the collection exists before proceeding
        logging.info("--- Getting/Creating collection... ---")
        try:
            collection = client.get_or_create_collection(name=THOUGHTS_COLLECTION, embedding_function=default_ef)
            logging.info(f"--- Collection '{collection.name}' obtained/created. ---")
        except Exception as e:
            logging.error(f"--- FAILED to get/create collection: {e} ---", exc_info=True)  # Root log error
            raise McpError(ErrorData(code=INTERNAL_ERROR, message=f"Could not access thinking collection: {str(e)}"))

        # Validate branch logic

        # Handle default values for session_id and branch_from_thought
        effective_session_id = session_id if session_id else str(uuid.uuid4())
        effective_branch_from_thought = branch_from_thought if branch_from_thought > 0 else None
        effective_branch_id = branch_id if branch_id else None
        timestamp = int(time.time())
        metadata = ThoughtMetadata(
            session_id=effective_session_id,
            thought_number=thought_number,
            total_thoughts=total_thoughts,
            timestamp=timestamp,
            branch_from_thought=effective_branch_from_thought,  # Use effective value
            branch_id=effective_branch_id,  # Use effective value
            next_thought_needed=next_thought_needed,
            custom_data=parsed_custom_data,
        )

        thought_id = f"thought_{effective_session_id}_{thought_number}"
        if effective_branch_id:  # Check effective value
            thought_id += f"_branch_{effective_branch_id}"

        # FIX: Construct metadata dictionary correctly using asdict for dataclass
        # Convert dataclass instance to dictionary
        base_metadata_dict = asdict(metadata)

        # Manually filter out None values
        filtered_metadata_dict = {k: v for k, v in base_metadata_dict.items() if v is not None}

        # Extract custom_data if it exists (and remove it from the main dict)
        # Use .pop on the filtered dict to get custom_data and remove it in one step
        custom_data_to_flatten = filtered_metadata_dict.pop("custom_data", None)

        # If custom_data was present and is a dict, flatten it with prefix
        if isinstance(custom_data_to_flatten, dict):
            for ck, cv in custom_data_to_flatten.items():
                filtered_metadata_dict[f"custom:{ck}"] = cv

        # The filtered_metadata_dict now contains the correct structure for Chroma
        metadata_dict_for_chroma = filtered_metadata_dict

        try:
            # Use root logger around add
            logging.info(f"--- Attempting collection.add for ID: {thought_id} ---")
            collection.add(documents=[thought], metadatas=[metadata_dict_for_chroma], ids=[thought_id])
            logging.info(f"--- Successfully added thought ID: {thought_id} ---")
        except (ValueError, InvalidDimensionException) as e:
            # Use root logger for error
            logging.error(f"--- FAILED collection.add: {e} ---", exc_info=True)
            raise McpError(
                ErrorData(
                    code=INTERNAL_ERROR,
                    message=f"ChromaDB Error adding thought: {str(e)}",
                )
            )

        # --- KEEPING Previous Thoughts Logic ---
        previous_thoughts = []
        if thought_number > 1:
            # Fetch ALL thoughts for the session first
            where_clause_get = {"session_id": effective_session_id}

            try:
                # Simpler get call, fetch all thoughts for the session
                results = collection.get(
                    where=where_clause_get,
                    include=["documents", "metadatas"],
                )

                thought_data = []
                if results and results.get("ids"):
                    for i in range(len(results["ids"])):
                        raw_meta = results["metadatas"][i] or {}
                        thought_num = raw_meta.get("thought_number")

                        # Filter for previous thoughts ONLY
                        if thought_num is None or thought_num >= thought_number:
                            continue  # Skip current/future thoughts

                        # --- Python-based branch filtering ---
                        thought_branch_id = raw_meta.get("branch_id")

                        if effective_branch_id:
                            # If we are IN a branch, include thoughts from this branch
                            # OR from the main trunk before the branch point
                            thought_branch_from = raw_meta.get("branch_from_thought")
                            is_in_correct_branch = thought_branch_id == effective_branch_id
                            is_on_main_trunk_before_branch = thought_branch_id is None and (
                                effective_branch_from_thought is None or thought_num < effective_branch_from_thought
                            )
                            if not (is_in_correct_branch or is_on_main_trunk_before_branch):
                                continue  # Skip thoughts not relevant to this branch history
                        else:
                            # If we are NOT in a branch, skip any thought that has a branch_id
                            if thought_branch_id is not None:
                                continue
                        # --- End Python-based branch filtering ---

                        # Reconstruct custom data (if any)
                        reconstructed_custom = {
                            k[len("custom:") :]: v for k, v in raw_meta.items() if k.startswith("custom:")
                        }
                        base_meta = {k: v for k, v in raw_meta.items() if not k.startswith("custom:")}
                        if reconstructed_custom:
                            base_meta["custom_data"] = reconstructed_custom

                        thought_data.append(
                            {
                                "id": results["ids"][i],
                                "content": results["documents"][i],
                                "metadata": base_meta,
                                "thought_number_sort_key": thought_num,  # Use the already retrieved thought_number
                            }
                        )

                # Sort based on thought_number
                sorted_thoughts = sorted(thought_data, key=lambda x: x["thought_number_sort_key"])

                # Final list without the sort key
                previous_thoughts = [
                    {k: v for k, v in thought.items() if k != "thought_number_sort_key"} for thought in sorted_thoughts
                ]

            except Exception as e:
                logger.warning(
                    f"Could not retrieve previous thoughts for session {effective_session_id}: {e}",
                    exc_info=True,
                )
        # --- END Previous Thoughts Logic ---

        # REMOVED: Call to non-existent _update_session_summary
        # if thought_number == total_thoughts and not next_thought_needed:
        #     await _update_session_summary(effective_session_id, collection, branch_id) # This function doesn't exist

        # Return List[TextContent] on success
        result_data = {
            "session_id": effective_session_id,
            "thought_id": thought_id,
            "previous_thoughts_count": len(previous_thoughts),
        }
        result_json = json.dumps(result_data, indent=2)

        logging.info("--- EXITING _base_sequential_thinking_impl NORMALLY ---")  # Root log exit
        return [types.TextContent(type="text", text=result_json)]

    except ValidationError as e:
        logger.error(f"Validation Error in sequential thinking: {e}", exc_info=True)
        # Raise McpError instead of returning CallToolResult
        raise McpError(ErrorData(code=INVALID_PARAMS, message=f"Invalid parameters: {str(e)}"))
    except McpError as e:  # Re-raise already caught McpError (from custom_data parsing)
        raise e
    except Exception as e:  # Catch other unexpected errors
        # Root log unexpected exceptions
        logging.error(
            f"--- EXITING _base_sequential_thinking_impl VIA EXCEPTION: {type(e).__name__} ---", exc_info=True
        )
        logger.error(f"Unexpected error during sequential thinking: {e}", exc_info=True)
        # Raise McpError instead of returning CallToolResult
        raise McpError(ErrorData(code=INTERNAL_ERROR, message=f"An unexpected error occurred: {str(e)}"))


async def _find_similar_thoughts_impl(input_data: FindSimilarThoughtsInput) -> List[types.TextContent]:
    """Performs a semantic search for similar thoughts.

    Args:
        input_data: A FindSimilarThoughtsInput object containing validated arguments.

    Returns:
        A list containing a single TextContent object with a JSON string representing
        the query results (similar to ChromaDB query results, including lists for ids,
        documents/thoughts, metadatas, distances).
        On error (e.g., invalid parameters, collection not found, unexpected issue),
        raises McpError.
    """

    logger = get_logger("tools.thinking")
    try:
        # Access validated data
        query = input_data.query
        session_id = input_data.session_id
        n_results = input_data.n_results  # Has default
        threshold = input_data.threshold
        include_branches = input_data.include_branches  # Has default

        # Use default threshold if -1.0 is passed
        effective_threshold = DEFAULT_SIMILARITY_THRESHOLD if threshold == -1.0 else threshold
        # Validate threshold range (0.0 to 1.0)
        if not (0.0 <= effective_threshold <= 1.0):
            raise McpError(
                ErrorData(
                    code=INVALID_PARAMS, message=f"Threshold must be between 0.0 and 1.0, got {effective_threshold}"
                )
            )

        client = get_chroma_client()

        # Get the server's default embedding function
        server_ef_name = get_server_config().embedding_function_name
        default_ef = get_embedding_function(server_ef_name)
        logger.debug(f"Using embedding function '{server_ef_name}' for {THOUGHTS_COLLECTION}.")

        # Get collection, handle not found specifically
        try:
            collection = client.get_collection(name=THOUGHTS_COLLECTION, embedding_function=default_ef)
        except ValueError as e:
            if f"Collection {THOUGHTS_COLLECTION} does not exist." in str(e):
                logger.warning(f"Cannot find similar thoughts: Collection '{THOUGHTS_COLLECTION}' not found.")
                # Return success with empty results, indicating collection doesn't exist
                return [  # Return list
                    types.TextContent(
                        type="text",
                        text=json.dumps(
                            {
                                "similar_thoughts": [],
                                "total_found": 0,
                                "threshold_used": effective_threshold,
                                "message": f"Collection '{THOUGHTS_COLLECTION}' not found.",
                            },
                            indent=2,
                        ),
                    )
                ]
            else:
                raise e  # Re-raise other ValueErrors
        except Exception as e:  # Catch other potential errors during get_collection
            logger.error(f"Error getting collection '{THOUGHTS_COLLECTION}' for query: {e}", exc_info=True)
            # Raise McpError instead of returning CallToolResult
            raise McpError(
                ErrorData(
                    code=INTERNAL_ERROR,
                    message=f"ChromaDB Error accessing collection '{THOUGHTS_COLLECTION}': {str(e)}",
                )
            )

        # Prepare where clause if session_id is provided
        where_clause = None
        if session_id:
            where_clause = {"session_id": session_id}
            # TODO: Add branch filtering logic if needed based on include_branches

        # Perform query, handle errors
        try:
            results = collection.query(
                query_texts=[query],
                n_results=n_results,
                where=where_clause,
                include=["documents", "metadatas", "distances"],
            )
        except ValueError as e:  # Catch query-specific errors
            logger.error(f"Error querying thoughts collection '{THOUGHTS_COLLECTION}': {e}", exc_info=True)
            # Raise McpError instead of returning CallToolResult
            raise McpError(
                ErrorData(
                    code=INTERNAL_ERROR,  # Use INTERNAL_ERROR for query issues
                    message=f"ChromaDB Query Error: {str(e)}",
                )
            )

        # Process results and filter by threshold
        similar_thoughts = []
        if results and results.get("ids") and results["ids"][0]:
            for i in range(len(results["ids"][0])):
                distance = results["distances"][0][i]
                similarity = 1.0 - distance  # Ensure float calculation

                if similarity >= effective_threshold:
                    raw_meta = results["metadatas"][0][i] or {}
                    # Reconstruct custom data
                    reconstructed_custom = {
                        k[len("custom:") :]: v for k, v in raw_meta.items() if k.startswith("custom:")
                    }
                    base_meta = {k: v for k, v in raw_meta.items() if not k.startswith("custom:")}
                    if reconstructed_custom:
                        base_meta["custom_data"] = reconstructed_custom

                    thought = {
                        "id": results["ids"][0][i],  # Include ID
                        "content": results["documents"][0][i],
                        "metadata": base_meta,
                        "similarity": similarity,
                    }
                    similar_thoughts.append(thought)

        # Success result
        result_data = {
            "similar_thoughts": similar_thoughts,
            "total_found": len(similar_thoughts),
            "threshold_used": effective_threshold,
        }
        result_json = json.dumps(result_data, indent=2)
        return [types.TextContent(type="text", text=result_json)]

    except ValueError as e:  # Catch ValueErrors re-raised from get_collection
        logger.error(f"Value error accessing collection '{THOUGHTS_COLLECTION}' for query: {e}", exc_info=False)
        # Raise McpError instead of returning CallToolResult
        raise McpError(
            ErrorData(
                code=INTERNAL_ERROR,  # Treat re-raised access error as internal
                message=f"ChromaDB Value Error accessing collection: {str(e)}",
            )
        )
    except Exception as e:
        logger.error(f"Unexpected error finding similar thoughts: {e}", exc_info=True)
        # Raise McpError instead of returning CallToolResult
        raise McpError(
            ErrorData(
                code=INTERNAL_ERROR,
                message=f"Tool Error: An unexpected error occurred while finding similar thoughts. Details: {str(e)}",
            )
        )


async def _get_session_summary_impl(input_data: GetSessionSummaryInput) -> List[types.TextContent]:
    """Fetches all thoughts belonging to a specific session.

    Args:
        input_data: A GetSessionSummaryInput object containing validated arguments.

    Returns:
        A list containing a single TextContent object with a JSON string containing
        a list of thoughts (documents/metadata), ordered sequentially by thought_number
        (and potentially by branch structure if included).
        On error (e.g., session not found, database error, unexpected issue),
        raises McpError.
    """

    logger = get_logger("tools.thinking")
    try:
        # Access validated data
        session_id = input_data.session_id
        include_branches = input_data.include_branches  # Has default

        client = get_chroma_client()

        # Get the server's default embedding function
        server_ef_name = get_server_config().embedding_function_name
        default_ef = get_embedding_function(server_ef_name)
        logger.debug(f"Using embedding function '{server_ef_name}' for {THOUGHTS_COLLECTION}.")

        # Ensure the collection exists before proceeding
        logger.debug(f"Ensuring collection '{THOUGHTS_COLLECTION}' exists for summary...")
        try:
            collection = client.get_or_create_collection(name=THOUGHTS_COLLECTION, embedding_function=default_ef)
            logger.debug(f"Collection '{THOUGHTS_COLLECTION}' obtained or created for summary.")
        except Exception as e:
            logger.error(f"Failed to get or create collection '{THOUGHTS_COLLECTION}' for summary: {e}", exc_info=True)
            raise McpError(ErrorData(code=INTERNAL_ERROR, message=f"Could not access thinking collection: {str(e)}"))

        # Get thoughts, handle errors
        try:
            # Fetch ALL documents first
            logger.debug(f"Fetching all documents from {THOUGHTS_COLLECTION}...")
            results = collection.get(include=["documents", "metadatas"])  # Fetch all
            logger.debug(f"Fetched {len(results.get('ids', []))} documents in total.")
        except ValueError as e:  # Catch errors from get (e.g., bad filter)
            logger.error(f"Error getting thoughts for session '{session_id}': {e}", exc_info=True)
            # Raise McpError instead of returning CallToolResult
            raise McpError(
                ErrorData(code=INTERNAL_ERROR, message=f"ChromaDB Get Error: {str(e)}")  # Treat get errors as internal
            )

        # Process and sort results
        session_thoughts = []
        if results and results.get("ids"):
            thought_data = []
            for i in range(len(results["ids"])):
                raw_meta = results["metadatas"][i] or {}
                # --- Filter by session_id in Python ---
                if raw_meta.get("session_id") != session_id:
                    continue
                # --- Filter by branch_id if needed ---
                if not include_branches and raw_meta.get("branch_id") is not None:
                    continue

                # Reconstruct custom data (Corrected logic)
                reconstructed_custom = {k[len("custom:") :]: v for k, v in raw_meta.items() if k.startswith("custom:")}
                base_meta = {k: v for k, v in raw_meta.items() if not k.startswith("custom:")}
                if reconstructed_custom:
                    base_meta["custom_data"] = reconstructed_custom

                thought_data.append(
                    {
                        "id": results["ids"][i],  # Include ID
                        "content": results["documents"][i],
                        "metadata": base_meta,  # Use the fully reconstructed base_meta
                        "thought_number_sort_key": base_meta.get("thought_number", 999999),  # Get from base_meta
                    }
                )

            # Sort based on thought_number
            sorted_thoughts = sorted(thought_data, key=lambda x: x["thought_number_sort_key"])

            # Final list without the sort key
            session_thoughts = [
                {k: v for k, v in thought.items() if k != "thought_number_sort_key"} for thought in sorted_thoughts
            ]

        # Success result - RESTORED
        result_data = {
            "session_id": session_id,
            "session_thoughts": session_thoughts,
            "total_thoughts_in_session": len(session_thoughts),
        }
        result_json = json.dumps(result_data, indent=2)
        return [types.TextContent(type="text", text=result_json)]

    except ValueError as e:  # Catch ValueErrors re-raised from get_collection
        logger.error(
            f"Value error accessing collection '{THOUGHTS_COLLECTION}' for session summary: {e}", exc_info=False
        )
        # Raise McpError instead of returning CallToolResult
        raise McpError(
            ErrorData(
                code=INTERNAL_ERROR,  # Treat re-raised access error as internal
                message=f"ChromaDB Value Error accessing collection: {str(e)}",
            )
        )
    except Exception as e:
        logger.error(f"Unexpected error getting session summary for '{session_id}': {e}", exc_info=True)
        # Raise McpError instead of returning CallToolResult
        raise McpError(
            ErrorData(
                code=INTERNAL_ERROR,
                message=f"Tool Error: An unexpected error occurred while getting session summary for '{session_id}'. Details: {str(e)}",
            )
        )


async def _find_similar_sessions_impl(input_data: FindSimilarSessionsInput) -> List[types.TextContent]:
    """Performs a semantic search for sessions similar to the query.

    (Note: This functionality might require pre-calculating session embeddings
     or performing aggregation queries, depending on the implementation.)

    Args:
        input_data: A FindSimilarSessionsInput object containing validated arguments.

    Returns:
        A list containing a single TextContent object with a JSON string listing
        similar session IDs and potentially their similarity scores.
        On error (e.g., supporting collection/index not found, invalid query,
        unexpected issue), raises McpError.
    """

    logger = get_logger("tools.thinking")
    try:
        # Access validated data
        query = input_data.query
        n_results = input_data.n_results  # Has default
        threshold = input_data.threshold

        # Use default threshold if -1.0 is passed
        effective_threshold = DEFAULT_SIMILARITY_THRESHOLD if threshold == -1.0 else threshold
        # Validate threshold range (0.0 to 1.0)
        if not (0.0 <= effective_threshold <= 1.0):
            raise McpError(
                ErrorData(
                    code=INVALID_PARAMS, message=f"Threshold must be between 0.0 and 1.0, got {effective_threshold}"
                )
            )

        client = get_chroma_client()

        # Get the server's default embedding function
        server_ef_name = get_server_config().embedding_function_name
        default_ef = get_embedding_function(server_ef_name)
        logger.debug(f"Using embedding function '{server_ef_name}' for thinking collections.")

        # --- Step 1: Get all unique session IDs from the thoughts collection ---
        thoughts_collection = None
        all_session_ids = set()
        try:
            thoughts_collection = client.get_collection(name=THOUGHTS_COLLECTION, embedding_function=default_ef)
            # Efficiently get all unique session_ids from metadata
            # This might be slow for very large collections, consider optimization if needed
            all_metadata = thoughts_collection.get(include=["metadatas"])
            if all_metadata and all_metadata.get("metadatas"):
                for meta in all_metadata["metadatas"]:
                    if meta and "session_id" in meta:
                        all_session_ids.add(meta["session_id"])
        except ValueError as e:
            if f"Collection {THOUGHTS_COLLECTION} does not exist." in str(e):
                logger.warning(f"Cannot find similar sessions: Collection '{THOUGHTS_COLLECTION}' not found.")
                # Return empty result if thoughts collection is missing
                return [  # Return list
                    types.TextContent(
                        type="text",
                        text=json.dumps(
                            {"similar_sessions": [], "total_found": 0, "threshold_used": effective_threshold}, indent=2
                        ),
                    )
                ]
            else:
                raise e  # Re-raise other ValueErrors
        except Exception as e:
            logger.error(f"Error accessing thoughts collection '{THOUGHTS_COLLECTION}': {e}", exc_info=True)
            # Raise McpError instead of returning CallToolResult
            raise McpError(
                ErrorData(code=INTERNAL_ERROR, message=f"ChromaDB Error accessing thoughts collection: {str(e)}")
            )

        if not all_session_ids:
            logger.info("No sessions found in the thoughts collection.")
            return [  # Return list
                types.TextContent(
                    type="text",
                    text=json.dumps(
                        {"similar_sessions": [], "total_found": 0, "threshold_used": effective_threshold}, indent=2
                    ),
                )
            ]

        # --- Step 2: Create/Get Sessions Collection and Embed Session Summaries ---
        sessions_collection = None
        try:
            # Try getting the sessions collection
            sessions_collection = client.get_or_create_collection(
                name=SESSIONS_COLLECTION, embedding_function=default_ef
            )
        except ValueError as e:
            # Handle case where SESSIONS_COLLECTION specifically does not exist
            if f"Collection {SESSIONS_COLLECTION} does not exist." in str(e):
                logger.warning(f"Cannot find similar sessions: Required collection '{SESSIONS_COLLECTION}' not found.")
                # Raise McpError instead of returning CallToolResult
                raise McpError(
                    ErrorData(
                        code=INTERNAL_ERROR,  # Use INTERNAL_ERROR as it's a required backend component
                        message=f"Tool Error: Collection '{SESSIONS_COLLECTION}' not found",
                    )
                )
            else:
                # Re-raise other ValueErrors to be caught by the general exception handler below
                raise e
        except Exception as e:
            # Catch other errors during get_collection (non-ValueError)
            logger.error(f"Error accessing sessions collection '{SESSIONS_COLLECTION}': {e}", exc_info=True)
            # Raise McpError instead of returning CallToolResult
            raise McpError(
                ErrorData(code=INTERNAL_ERROR, message=f"ChromaDB Error accessing sessions collection: {str(e)}")
            )

        # If collection exists, proceed with embedding and adding summaries
        try:
            # Embed summaries for sessions not already in the sessions collection
            existing_session_ids = set(sessions_collection.get().get("ids", []))
            sessions_to_embed = []
            ids_to_embed = []

            for session_id in all_session_ids:
                if session_id not in existing_session_ids:
                    try:
                        # Get session summary (returns List[TextContent] on success)
                        summary_result_list = await _get_session_summary_impl(
                            GetSessionSummaryInput(session_id=session_id)
                        )
                        # Safely parse the JSON content from the first TextContent item
                        summary_data = json.loads(summary_result_list[0].text)
                        summary_text = " ".join(
                            [t.get("content", "") for t in summary_data.get("session_thoughts", [])]
                        )
                        if summary_text:  # Only embed if there's content
                            logger.debug(
                                f"Generated summary for session '{session_id}': '{summary_text[:100]}...'"
                            )  # Log summary
                            sessions_to_embed.append(summary_text)
                            ids_to_embed.append(session_id)
                    except McpError as summary_error:
                        logger.warning(
                            f"Failed to get summary for session '{session_id}' to embed: {summary_error.message}"
                        )
                    except (json.JSONDecodeError, IndexError, AttributeError) as parse_error:
                        logger.warning(f"Could not parse summary result for session '{session_id}': {parse_error}")

            if sessions_to_embed:
                logger.info(f"Embedding summaries for {len(sessions_to_embed)} new/updated sessions.")
                logger.debug(f"IDs to embed: {ids_to_embed}")  # Log IDs before add
                logger.debug(f"Summaries to embed: {sessions_to_embed}")  # Log summaries before add
                sessions_collection.add(documents=sessions_to_embed, ids=ids_to_embed)
                logger.info(f"Finished adding/embedding summaries to '{SESSIONS_COLLECTION}'.")  # Log after add

        except Exception as e:
            # Catch errors during the embedding/adding process
            logger.error(f"Error embedding/adding to sessions collection '{SESSIONS_COLLECTION}': {e}", exc_info=True)
            # Raise McpError instead of returning CallToolResult
            raise McpError(ErrorData(code=INTERNAL_ERROR, message=f"ChromaDB Error updating sessions: {str(e)}"))

        # --- Step 3: Query the Sessions Collection ---
        similar_sessions = []
        if sessions_collection:  # Ensure collection was accessed successfully
            try:
                query_results = sessions_collection.query(
                    query_texts=[query],
                    n_results=n_results,
                    include=["metadatas", "distances"],  # Only need distance and ID (implicit)
                )

                if query_results and query_results.get("ids") and query_results["ids"][0]:
                    for i in range(len(query_results["ids"][0])):
                        session_id = query_results["ids"][0][i]
                        distance = query_results["distances"][0][i]
                        similarity = 1.0 - distance

                        if similarity >= effective_threshold:
                            try:
                                # Fetch the full summary details again for the result
                                summary_result_list = await _get_session_summary_impl(
                                    GetSessionSummaryInput(session_id=session_id)
                                )
                                # Safely parse the JSON content
                                summary_data = json.loads(summary_result_list[0].text)
                                summary_data["similarity_score"] = similarity  # Add score
                                similar_sessions.append(summary_data)
                            except McpError as summary_error:
                                logger.warning(
                                    f"Failed to get final summary for session '{session_id}': {summary_error.message}"
                                )
                            except (json.JSONDecodeError, IndexError, AttributeError) as parse_error:
                                logger.warning(
                                    f"Could not parse final summary result for session '{session_id}': {parse_error}"
                                )
            except ValueError as e:
                logger.error(f"Error querying sessions collection '{SESSIONS_COLLECTION}': {e}", exc_info=True)
                # Raise McpError instead of returning CallToolResult
                raise McpError(ErrorData(code=INTERNAL_ERROR, message=f"ChromaDB Query Error on sessions: {str(e)}"))

        # Success result
        result_data = {
            "similar_sessions": similar_sessions,
            "total_found": len(similar_sessions),
            "threshold_used": effective_threshold,
        }
        result_json = json.dumps(result_data, indent=2)
        return [types.TextContent(type="text", text=result_json)]

    except ValueError as e:  # Catch ValueErrors re-raised from get_collection (thoughts)
        logger.error(f"Value error accessing thoughts collection '{THOUGHTS_COLLECTION}': {e}", exc_info=False)
        # Raise McpError instead of returning CallToolResult
        raise McpError(
            ErrorData(
                code=INTERNAL_ERROR,  # Treat re-raised access error as internal
                message=f"ChromaDB Value Error accessing thoughts collection: {str(e)}",
            )
        )
    except Exception as e:
        logger.error(f"Unexpected error finding similar sessions: {e}", exc_info=True)
        # Raise McpError instead of returning CallToolResult
        raise McpError(
            ErrorData(
                code=INTERNAL_ERROR,
                message=f"Tool Error: An unexpected error occurred while finding similar sessions. Details: {str(e)}",
            )
        )


# Ensure mcp instance is imported/available for decorators
